from numpy import NaN, isnan

from ._identifiers import IDENTIFIERS
from ..helpers import BaseSectionObject
from .._type_converter import to_bool
from ..section_labels import CONDUITS, ORIFICES, OUTLETS, PUMPS, WEIRS


class _Link(BaseSectionObject):
    _identifier = IDENTIFIERS.Name

    def __init__(self, Name, FromNode, ToNode):
        self.Name = str(Name)
        self.FromNode = str(FromNode)
        self.ToNode = str(ToNode)


class Conduit(_Link):
    """Section: **[CONDUITS]**

    Purpose:
        Identifies each conduit link of the drainage system. Conduits are pipes or channels that convey water
        from one node to another.

    Format:
        ::

            Name  Node1  Node2  Length  N  Z1  Z2  (Q0  Qmax)

    Format-PCSWMM:
        ``Name InletNode OutletNode Length ManningN InletOffset OutletOffset InitFlow MaxFlow``

    Format-SWMM-GUI:
        ``Name FromNode ToNode Length Roughness InOffset OutOffset InitFlow MaxFlow``

    Remarks:
        These offsets are expressed as a relative distance above the node invert if the ``LINK_OFFSETS`` option
        is set to ``DEPTH`` (the default) or as an absolute elevation if it is set to ``ELEVATION``.

    Args:
        Name (str): name assigned to conduit link.
        FromNode (str): name of upstream node. ``Node1``
        ToNode (str): name of downstream node. ``Node2``
        Length (float): conduit length (ft or m).
        Roughness (float): value of n (i.e., roughness parameter) in Manning’s equation. ``N``
        InOffset (float): offset of upstream end of conduit invert above the invert elevation of its
                          upstream node (ft or m). ``Z1``
        OutOffset (float): offset of downstream end of conduit invert above the invert elevation of its
                          downstream node (ft or m). ``Z2``
        InitFlow (float): flow in conduit at start of simulation (flow units) (default is 0). ``Q0``
        MaxFlow (float): maximum flow allowed in the conduit (flow units) (default is no limit | 0 = no limit). ``Qmax``

    Attributes:
        Name (str): name assigned to conduit link.
        FromNode (str): name of upstream node. ``Node1``
        ToNode (str): name of downstream node. ``Node2``
        Length (float): conduit length (ft or m).
        Roughness (float): value of n (i.e., roughness parameter) in Manning’s equation. ``N``
        InOffset (float): offset of upstream end of conduit invert above the invert elevation of its
                          upstream node (ft or m). ``Z1``
        OutOffset (float): offset of downstream end of conduit invert above the invert elevation of its
                          downstream node (ft or m). ``Z2``
        InitFlow (float): flow in conduit at start of simulation (flow units) (default is 0). ``Q0``
        MaxFlow (float): maximum flow allowed in the conduit (flow units) (default is no limit | 0 = no limit). ``Qmax``
    """
    _section_label = CONDUITS

    def __init__(self, Name, FromNode, ToNode, Length, Roughness, InOffset=0, OutOffset=0, InitFlow=0, MaxFlow=NaN):
        _Link.__init__(self, Name, FromNode, ToNode)
        self.Length = float(Length)
        self.Roughness = float(Roughness)
        self.InOffset = float(InOffset)
        self.OutOffset = float(OutOffset)
        self.InitFlow = float(InitFlow)
        self.MaxFlow = float(MaxFlow)
        if self.MaxFlow == 0:
            self.MaxFlow = NaN


class Orifice(_Link):
    """
    Section: [**ORIFICES**]

    Purpose:
        Identifies each orifice link of the drainage system. An orifice link serves to limit the
        flow exiting a node and is often used to model flow diversions and storage node
        outlets.

    Format:
        ::

            Name Node1 Node2 Type Offset Cd (Flap Orate)

    Format-PCSWMM:
        ``Name  InletNode  OutletNode  OrificeType  CrestHeight  DischCoeff  FlapGate  Open/CloseTime``

    Format-SWMM-GUI:
        ``Name  FromNode  ToNode  Type  Offset  Qcoeff  Gated  CloseTime``

    Remarks:
        The geometry of an orifice’s opening must be described in the [``XSECTIONS``] section.
        The only allowable shapes are CIRCULAR and RECT_CLOSED (closed rectangular).

    Args:
        Name (str): name assigned to orifice link.
        FromNode (str): name of node on inlet end of orifice. ``Node1``
        ToNode (str): name of node on outlet end of orifice. ``Node2``
        Type (str): orientation of orifice: either SIDE or BOTTOM.
        Offset (float): amount that a Side Orifice’s bottom or the position of a Bottom Orifice is offset above
            the invert of inlet node (ft or m, expressed as either a depth or as an elevation,
            depending on the LINK_OFFSETS option setting).
        Qcoeff (float): discharge coefficient (unitless). ``Cd``
        FlapGate (bool): YES if flap gate present to prevent reverse flow, NO if not (default is NO). ``Flap``
        Orate (int): time in decimal hours to open a fully closed orifice (or close a fully open one).
                        Use 0 if the orifice can open/close instantaneously.

    Attributes:
        Name (str): name assigned to orifice link.
        FromNode (str): name of node on inlet end of orifice. ``Node1``
        ToNode (str): name of node on outlet end of orifice. ``Node2``
        Type (str): orientation of orifice: either SIDE or BOTTOM.
        Offset (float): amount that a Side Orifice’s bottom or the position of a Bottom Orifice is offset above
            the invert of inlet node (ft or m, expressed as either a depth or as an elevation,
            depending on the LINK_OFFSETS option setting).
        Qcoeff (float): discharge coefficient (unitless). ``Cd``
        FlapGate (bool): YES if flap gate present to prevent reverse flow, NO if not (default is NO). ``Flap``
        Orate (int): time in decimal hours to open a fully closed orifice (or close a fully open one).
                        Use 0 if the orifice can open/close instantaneously.
    """
    _section_label = ORIFICES

    class TYPES:
        """orientation of orifice: either SIDE or BOTTOM"""
        SIDE = 'SIDE'
        BOTTOM = 'BOTTOM'

    def __init__(self, Name, FromNode, ToNode, Type, Offset, Qcoeff, FlapGate=False, Orate=0):
        _Link.__init__(self, Name, FromNode, ToNode)
        self.Type = str(Type)
        self.Offset = float(Offset)
        self.Qcoeff = float(Qcoeff)
        self.FlapGate = to_bool(FlapGate)
        self.Orate = int(Orate)


class Outlet(_Link):
    """
    Section: [**OUTLETS**]

    Purpose:
        Identifies each outlet flow control device of the drainage system.
        These devices are used to model outflows from storage units or flow diversions
        that have a user-defined relation between flow rate and water depth.

    Formats:
        ::

            Name Node1 Node2 Offset TABULAR/DEPTH Qcurve (Gated)
            Name Node1 Node2 Offset TABULAR/HEAD Qcurve (Gated)
            Name Node1 Node2 Offset FUNCTIONAL/DEPTH C1 C2 (Gated)
            Name Node1 Node2 Offset FUNCTIONAL/HEAD C1 C2 (Gated)

    Format-PCSWMM:
        ``Name InletNode OutletNode OutflowHeight OutletType Qcoeff/QTable Qexpon FlapGate``

    Format-SWMM-GUI:
        ``Name FromNode ToNode Offset Type QTable/Qcoeff Qexpon Gated``

    Args:
        Name (str): name assigned to outlet link.
        FromNode (str): name of node on inlet end of link. ``Node1``
        ToNode (str): name of node on outflow end of link. ``Node2``
        Offset (float): amount that the outlet is offset above the invert of inlet node
            (ft or m, expressed as either a depth or as an elevation, depending on the LINK_OFFSETS option setting).
        Type (str): one of Outlet.Types
        *args: for automatic input file reading
        Curve (str | tuple[float, float]):

            - :obj:`float`: ``Qcurve`` name of the rating curve listed in the [``CURVES``] section that describes outflow rate (flow units) as a function of:

                    - water depth above the offset elevation at the inlet node (ft or m) for a TABULAR/DEPTH outlet
                    - head difference (ft or m) between the inlet and outflow nodes for a TABULAR/HEAD outlet.
            - :obj:`tuple[float, float]`: (``C1``, ``C2``) coefficient and exponent, respectively, of a power function that relates outflow (Q) to:

                    - water depth (ft or m) above the offset elevation at the inlet node for a FUNCTIONAL/DEPTH outlet
                    - head difference (ft or m) between the inlet and outflow nodes for a FUNCTIONAL/HEAD outlet.
                      (i.e., Q = C1H^C2 where H is either depth or head).

        Gated (bool): ``YES`` if flap gate present to prevent reverse flow, ``NO`` if not (default is ``NO``).

    Attributes:
        Name (str): name assigned to outlet link.
        FromNode (str): name of node on inlet end of link. ``Node1``
        ToNode (str): name of node on outflow end of link. ``Node2``
        Offset (float): amount that the outlet is offset above the invert of inlet node
            (ft or m, expressed as either a depth or as an elevation, depending on the LINK_OFFSETS option setting).
        Type (str): one of Outlet.Types
        Curve (str | tuple[float, float]):

            - :obj:`float`: ``Qcurve`` name of the rating curve listed in the [``CURVES``] section that describes outflow rate (flow units) as a function of:

                    - water depth above the offset elevation at the inlet node (ft or m) for a TABULAR/DEPTH outlet
                    - head difference (ft or m) between the inlet and outflow nodes for a TABULAR/HEAD outlet.
            - :obj:`tuple[float, float]`: (``C1``, ``C2``) coefficient and exponent, respectively, of a power function that relates outflow (Q) to:

                    - water depth (ft or m) above the offset elevation at the inlet node for a FUNCTIONAL/DEPTH outlet
                    - head difference (ft or m) between the inlet and outflow nodes for a FUNCTIONAL/HEAD outlet.
                      (i.e., Q = C1H^C2 where H is either depth or head).

        Gated (bool): ``YES`` if flap gate present to prevent reverse flow, ``NO`` if not (default is ``NO``).
    """
    _section_label = OUTLETS

    class TYPES:
        TABULAR_DEPTH = 'TABULAR/DEPTH'
        TABULAR_HEAD = 'TABULAR/HEAD'
        FUNCTIONAL_DEPTH = 'FUNCTIONAL/DEPTH'
        FUNCTIONAL_HEAD = 'FUNCTIONAL/HEAD'

    def __init__(self, Name, FromNode, ToNode, Offset, Type, *args, Curve=None, Gated=False):
        _Link.__init__(self, Name, FromNode, ToNode)
        self.Offset = float(Offset)
        self.Type = str(Type)

        if args:
            if Type.startswith('TABULAR'):
                self._tabular_init(*args)

            elif Type.startswith('FUNCTIONAL'):
                self._functional_init(*args)

            else:
                raise NotImplementedError('Type: "{}" is not implemented'.format(Type))

        else:
            self.Curve = Curve
            self.Gated = to_bool(Gated)

    def _tabular_init(self, Qcurve, Gated=False):
        self.Curve = str(Qcurve)
        self.Gated = to_bool(Gated)

    def _functional_init(self, C1, C2, Gated=False):
        self.Curve = [float(C1), float(C2)]
        self.Gated = to_bool(Gated)


class Pump(_Link):
    """
    Section: [**PUMPS**]

    Purpose:
        Identifies each pump link of the drainage system.

    Format:
        ::

            Name Node1 Node2 Pcurve (Status Startup Shutoff)

    Format-PCSWMM:
        ``Name  InletNode  OutletNode  PumpCurve  InitStatus  StartupDepth  ShutoffDepth``

    Format-SWMM-GUI:
        ``Name  FromNode  ToNode  PumpCurve  Status  Sartup  Shutoff``

    Args:
        Name (str): name assigned to pump link.
        FromNode (str): name of node on inlet side of pump. ``Node1``
        ToNode (str): name of node on outlet side of pump. ``Node2``
        Curve (str): name of pump curve listed in the [CURVES] section of the input. ``Pcurve``
        Status (str): status at start of simulation (either ON or OFF; default is ON).
        Startup (float): depth at inlet node when pump turns on (ft or m) (default is 0).
        Shutoff (float): depth at inlet node when pump shuts off (ft or m) (default is 0).

    Attributes:
        Name (str): name assigned to pump link.
        FromNode (str): name of node on inlet side of pump. ``Node1``
        ToNode (str): name of node on outlet side of pump. ``Node2``
        Curve (str): name of pump curve listed in the [CURVES] section of the input. ``Pcurve``
        Status (str): status at start of simulation (either ON or OFF; default is ON).
        Startup (float): depth at inlet node when pump turns on (ft or m) (default is 0).
        Shutoff (float): depth at inlet node when pump shuts off (ft or m) (default is 0).

    See Section 3.2 for a description of the different types of pumps available.
    """
    _section_label = PUMPS

    class STATES:
        """status at start of simulation (either ON or OFF; default is ON)."""
        ON = 'ON'
        OFF = 'OFF'

    def __init__(self, Name, FromNode, ToNode, Curve, Status='ON', Startup=0, Shutoff=0):
        _Link.__init__(self, Name, FromNode, ToNode)
        self.Curve = str(Curve)
        self.Status = str(Status)
        self.Startup = float(Startup)
        self.Shutoff = float(Shutoff)


class Weir(_Link):
    """
    Section: [**WEIRS**]

    Purpose:
        Identifies each weir link of the drainage system. Weirs are used to model flow
        diversions and storage node outlets.

    Format:
        ::

            Name Node1 Node2 Type CrestHt Cd (Gated EC Cd2 Sur (Width Surface))

    Format-PCSWMM:
        ``Name InletNode OutletNode WeirType CrestHeight DischCoeff FlapGate EndCon EndCoeff Surcharge RoadWidth RoadSurf``

    Format-SWMM-GUI:
        ``Name FromNode ToNode Type CrestHt Qcoeff Gated EndCon EndCoeff Surcharge RoadWidth RoadSurf``

    Remarks:
        The geometry of a weir’s opening is described in the [XSECTIONS] section.
        The following shapes must be used with each type of weir:

        =============  ===================
        Weir Type      Cross-Section Shape
        =============  ===================
        Transverse     ``RECT_OPEN``
        Sideflow       ``RECT_OPEN``
        V-Notch        ``TRIANGULAR``
        Trapezoidal    ``TRAPEZOIDAL``
        Roadway        ``RECT_OPEN``
        =============  ===================

        The ``ROADWAY`` weir is a broad crested rectangular weir used model roadway crossings usually in conjunction with
        culvert-type conduits. It uses the FHWA HDS-5 method to determine a discharge coefficient as a function of flow
        depth and roadway width and surface.
        If no roadway data are provided then the weir behaves as a ``TRANSVERSE`` weir with Cd as its discharge coefficient.
        Note that if roadway data are provided, then values for the other optional weir parameters
        (``NO`` for Gated, 0 for EC, 0 for Cd2, and ``NO`` for Sur)
        must be entered even though they do not apply to ``ROADWAY`` weirs.

    Args:
        Name (str): name assigned to weir link
        FromNode (str): name of node on inlet side of weir. ``Node1``
        ToNode (str): name of node on outlet side of weir. ``Node2``
        Type (str): ``TRANSVERSE``, ``SIDEFLOW``, ``V-NOTCH``, ``TRAPEZOIDAL`` or ``ROADWAY``.
        CrestHeight (float): amount that the weir’s crest is offset above the invert of inlet node (ft or m,
            expressed as either a depth or as an elevation, depending on the LINK_OFFSETS option setting). ``CrestHt``
        Qcoeff (float): weir discharge coefficient (for CFS if using US flow units or CMS if using metric flow units). ``Cd``
        FlapGate (bool): ``YES`` if flap gate present to prevent reverse flow, ``NO`` if not (default is ``NO``). ``Gated``
        EndContractions (float): number of end contractions for ``TRANSVERSE`` or ``TRAPEZOIDAL`` weir (default is 0). ``EC``
        EndCoeff (float): discharge coefficient for triangular ends of a ``TRAPEZOIDAL`` weir
                         (for ``CFS`` if using US flow units or ``CMS`` if using metric flow units)
                         (default is value of Cd). ``Cd2``
        Surcharge (bool): ``YES`` if the weir can surcharge
            (have an upstream water level higher than the height of the opening);
            ``NO`` if it cannot (default is ``YES``). ``Sur``
        RoadWidth (float): width of road lanes and shoulders for ``ROADWAY`` weir (ft or m). ``Width``
        RoadSurface (str): type of road surface for ``ROADWAY`` weir: ``PAVED`` or ``GRAVEL``. ``Surface``

    Attributes:
        Name (str): name assigned to weir link
        FromNode (str): name of node on inlet side of weir. ``Node1``
        ToNode (str): name of node on outlet side of weir. ``Node2``
        Type (str): ``TRANSVERSE``, ``SIDEFLOW``, ``V-NOTCH``, ``TRAPEZOIDAL`` or ``ROADWAY``.
        CrestHeight (float): amount that the weir’s crest is offset above the invert of inlet node (ft or m,
            expressed as either a depth or as an elevation, depending on the LINK_OFFSETS option setting). ``CrestHt``
        Qcoeff (float): weir discharge coefficient (for CFS if using US flow units or CMS if using metric flow units). ``Cd``
        FlapGate (bool): ``YES`` if flap gate present to prevent reverse flow, ``NO`` if not (default is ``NO``). ``Gated``
        EndContractions (float): number of end contractions for ``TRANSVERSE`` or ``TRAPEZOIDAL`` weir (default is 0). ``EC``
        EndCoeff (float): discharge coefficient for triangular ends of a ``TRAPEZOIDAL`` weir
                         (for ``CFS`` if using US flow units or ``CMS`` if using metric flow units)
                         (default is value of Cd). ``Cd2``
        Surcharge (bool): ``YES`` if the weir can surcharge
            (have an upstream water level higher than the height of the opening);
            ``NO`` if it cannot (default is ``YES``). ``Sur``
        RoadWidth (float): width of road lanes and shoulders for ``ROADWAY`` weir (ft or m). ``Width``
        RoadSurface (str): type of road surface for ``ROADWAY`` weir: ``PAVED`` or ``GRAVEL``. ``Surface``
    """
    _section_label = WEIRS

    class TYPES:
        TRANSVERSE = 'TRANSVERSE'
        SIDEFLOW = 'SIDEFLOW'
        V_NOTCH = 'V-NOTCH'
        TRAPEZOIDAL = 'TRAPEZOIDAL'
        ROADWAY = 'ROADWAY'

    class ROADSURFACES:
        PAVED = 'PAVED'
        GRAVEL = 'GRAVEL'

    def __init__(self, Name, FromNode, ToNode, Type, CrestHeight, Qcoeff, FlapGate=False, EndContractions=0,
                 EndCoeff=NaN,
                 Surcharge=True, RoadWidth=NaN, RoadSurface=NaN):
        _Link.__init__(self, Name, FromNode, ToNode)
        self.Type = str(Type)
        self.CrestHeight = float(CrestHeight)
        self.Qcoeff = float(Qcoeff)
        self.FlapGate = to_bool(FlapGate)
        self.EndContractions = EndContractions
        if not isinstance(EndCoeff, str) and isnan(EndCoeff):
            EndCoeff = Qcoeff
        self.EndCoeff = float(EndCoeff)
        self.Surcharge = to_bool(Surcharge)
        self.RoadWidth = float(RoadWidth)
        self.RoadSurface = RoadSurface



from .collection import links_dict
from .macros import find_link
from ..misc.curve_simplification import _vec2d_dist, ramer_douglas
from ..sections import Vertices, Coordinate, Polygon
from ..section_labels import COORDINATES, VERTICES, POLYGONS


def transform_coordinates(inp, from_proj='epsg:31256', to_proj='epsg:32633'):
    """
    transform all coordinates of the sections COORDINATES, VERTICES and POLYGONS from one projection to another.

    Args:
        inp (swmm_api.SwmmInput): SWMM input data
        from_proj (str): Projection of data
        to_proj (str): Projection for resulting data

    .. Important::
        works inplace
    """
    from pyproj import Transformer
    # GK M34 to WGS 84 UTM zone 33N
    transformer = Transformer.from_crs(from_proj, to_proj, always_xy=True)
    # -----------------------------------
    if COORDINATES in inp:
        for node in inp[COORDINATES]:
            c = inp[COORDINATES][node]  # type: Coordinate
            c.x, c.y = transformer.transform(c.x, c.y)

    if VERTICES in inp:
        for link in inp[VERTICES]:
            v = inp[VERTICES][link]  # type: Vertices
            x, y = list(zip(*v.vertices))
            v.vertices = list(zip(*transformer.transform(x, y)))

    if POLYGONS in inp:
        for subcatchment in inp[POLYGONS]:
            p = inp[POLYGONS][subcatchment]  # type: Polygon
            x, y = list(zip(*p.polygon))
            p.polygon = list(zip(*transformer.transform(x, y)))


def complete_link_vertices(inp, label_link):
    """
    add node coordinates to vertices of a single link (start and end point)

    .. Important::
        works inplace

    Args:
        inp (swmm_api.SwmmInput): SWMM input data
        label_link (str): label of the link
    """
    link = find_link(inp, label_link)
    inp[VERTICES][label_link].vertices = ([inp[COORDINATES][link.FromNode].point] +
                                          inp[VERTICES][label_link].vertices +
                                          [inp[COORDINATES][link.ToNode].point])


def complete_vertices(inp):
    """
    add node coordinates to vertices of all links (start and end point)

    important for GIS export or GIS operations

    .. Important::
        works inplace

    Args:
        inp (swmm_api.SwmmInput): SWMM input data
    """
    if COORDINATES in inp:
        links = links_dict(inp)
        if links:
            if VERTICES not in inp:
                inp[VERTICES] = Vertices.create_section()

            object_type = inp[VERTICES]._section_object

            for label_link in links:  # type: Conduit # or Weir or Orifice or Pump or Outlet
                if label_link not in inp[VERTICES]:
                    inp[VERTICES].add_obj(object_type(label_link, vertices=list()))
                complete_link_vertices(inp, label_link)


def reduce_vertices(inp, node_range=0.25):
    """
    remove first and last vertices to keep only inner vertices (SWMM standard)

    important if data originally from GIS and export to SWMM

    Args:
        inp (swmm_api.SwmmInput): SWMM input data
        node_range (float): minimal distance in m from the first and last vertices to the end nodes

    .. Important::
        works inplace
    """
    links = links_dict(inp)

    for link_label in inp.VERTICES:
        l = links[link_label]
        v = inp[VERTICES][link_label].vertices
        p = inp[COORDINATES][l.FromNode].point
        if _vec2d_dist(p, v[0]) < node_range:
            v = v[1:]

        if v:
            p = inp[COORDINATES][l.ToNode].point
            if _vec2d_dist(p, v[-1]) < node_range:
                v = v[:-1]

        if v:
            inp[VERTICES][link_label].vertices = v
        else:
            del inp[VERTICES][link_label]


def simplify_link_vertices(vertices, dist=1.):
    """
    removes unneeded vertices with the Ramer-Douglas simplification

    Args:
        vertices (Vertices): vertices-object of link
        dist (float): threshold of algorythm

    .. Important::
        works inplace
    """
    vertices.vertices = ramer_douglas(vertices.vertices, dist)


def simplify_vertices(inp, dist=1.):
    """
    removes unneeded vertices with the Ramer-Douglas simplification

    Args:
        inp (swmm_api.SwmmInput): SWMM input data
        dist (float):  threshold of algorythm

    .. Important::
        works inplace
    """
    for v in inp.VERTICES:
        simplify_link_vertices(inp.VERTICES[v], dist)



ERROR_CODE = {101: "\n  ERROR 101: memory allocation error.",
              103: "\n  ERROR 103: cannot solve KW equations for Link %s.",
              105: "\n  ERROR 105: cannot open ODE solver.",
              107: "\n  ERROR 107: cannot compute a valid time step.",

              108: "\n  ERROR 108: ambiguous outlet ID name for Subcatchment %s.",
              109: "\n  ERROR 109: invalid parameter values for Aquifer %s.",
              110: "\n  ERROR 110: ground elevation is below water table for Subcatchment %s.",

              111: "\n  ERROR 111: invalid length for Conduit %s.",
              112: "\n  ERROR 112: elevation drop exceeds length for Conduit %s.",
              113: "\n  ERROR 113: invalid roughness for Conduit %s.",
              114: "\n  ERROR 114: invalid number of barrels for Conduit %s.",
              115: "\n  ERROR 115: adverse slope for Conduit %s.",
              117: "\n  ERROR 117: no cross section defined for Link %s.",
              119: "\n  ERROR 119: invalid cross section for Link %s.",
              121: "\n  ERROR 121: missing or invalid pump curve assigned to Pump %s.",
              122: "\n  ERROR 122: startup depth not higher than shutoff depth for Pump %s.",

              131: "\n  ERROR 131: the following links form cyclic loops in the drainage system:",
              133: "\n  ERROR 133: Node %s has more than one outlet link.",
              134: "\n  ERROR 134: Node %s has illegal DUMMY link connections.",

              135: "\n  ERROR 135: Divider %s does not have two outlet links.",
              136: "\n  ERROR 136: Divider %s has invalid diversion link.",
              137: "\n  ERROR 137: Weir Divider %s has invalid parameters.",
              138: "\n  ERROR 138: Node %s has initial depth greater than maximum depth.",
              139: "\n  ERROR 139: Regulator %s is the outlet of a non-storage node.",
              140: "\n  ERROR 140: Storage node %s has negative volume at full depth.",  # (5.1.015)
              141: "\n  ERROR 141: Outfall %s has more than 1 inlet link or an outlet link.",
              143: "\n  ERROR 143: Regulator %s has invalid cross-section shape.",
              145: "\n  ERROR 145: Drainage system has no acceptable outlet nodes.",

              151: "\n  ERROR 151: a Unit Hydrograph in set %s has invalid time base.",
              153: "\n  ERROR 153: a Unit Hydrograph in set %s has invalid response ratios.",
              155: "\n  ERROR 155: invalid sewer area for RDII at node %s.",

              156: "\n  ERROR 156: ambiguous station ID for Rain Gage %s.",
              157: "\n  ERROR 157: inconsistent rainfall format for Rain Gage %s.",
              158: "\n  ERROR 158: time series for Rain Gage %s is also used by another object.",
              159: "\n  ERROR 159: recording interval greater than time series interval for Rain Gage %s.",

              161: "\n  ERROR 161: cyclic dependency in treatment functions at node %s.",

              171: "\n  ERROR 171: Curve %s has invalid or out of sequence data.",
              173: "\n  ERROR 173: Time Series %s has its data out of sequence.",

              181: "\n  ERROR 181: invalid Snow Melt Climatology parameters.",
              182: "\n  ERROR 182: invalid parameters for Snow Pack %s.",

              183: "\n  ERROR 183: no type specified for LID %s.",
              184: "\n  ERROR 184: missing layer for LID %s.",
              185: "\n  ERROR 185: invalid parameter value for LID %s.",
              186: "\n  ERROR 186: invalid parameter value for LID placed in Subcatchment %s.",
              187: "\n  ERROR 187: LID area exceeds total area for Subcatchment %s.",
              188: "\n  ERROR 188: LID capture area exceeds total impervious area for Subcatchment %s.",

              191: "\n  ERROR 191: simulation start date comes after ending date.",
              193: "\n  ERROR 193: report start date comes after ending date.",
              195: "\n  ERROR 195: reporting time step or duration is less than routing time step.",

              200: "\n  ERROR 200: one or more errors in input file.",
              201: "\n  ERROR 201: too many characters in input line ",
              203: "\n  ERROR 203: too few items ",
              205: "\n  ERROR 205: invalid keyword %s ",
              207: "\n  ERROR 207: duplicate ID name %s ",
              209: "\n  ERROR 209: undefined object %s ",
              211: "\n  ERROR 211: invalid number %s ",
              213: "\n  ERROR 213: invalid date/time %s ",
              217: "\n  ERROR 217: control rule clause invalid or out of sequence ",  # (5.1.008)
              219: "\n  ERROR 219: data provided for unidentified transect ",
              221: "\n  ERROR 221: transect station out of sequence ",
              223: "\n  ERROR 223: Transect %s has too few stations.",
              225: "\n  ERROR 225: Transect %s has too many stations.",
              227: "\n  ERROR 227: Transect %s has no Manning's N.",
              229: "\n  ERROR 229: Transect %s has invalid overbank locations.",
              231: "\n  ERROR 231: Transect %s has no depth.",
              233: "\n  ERROR 233: invalid treatment function expression ",

              301: "\n  ERROR 301: files share same names.",
              303: "\n  ERROR 303: cannot open input file.",
              305: "\n  ERROR 305: cannot open report file.",
              307: "\n  ERROR 307: cannot open binary results file.",
              309: "\n  ERROR 309: error writing to binary results file.",
              311: "\n  ERROR 311: error reading from binary results file.",

              313: "\n  ERROR 313: cannot open scratch rainfall interface file.",
              315: "\n  ERROR 315: cannot open rainfall interface file %s.",
              317: "\n  ERROR 317: cannot open rainfall data file %s.",
              318: "\n  ERROR 318: the following line is out of sequence in rainfall data file %s.",  # (5.1.010)
              319: "\n  ERROR 319: unknown format for rainfall data file %s.",
              320: "\n  ERROR 320: invalid format for rainfall interface file.",
              321: "\n  ERROR 321: no data in rainfall interface file for gage %s.",

              323: "\n  ERROR 323: cannot open runoff interface file %s.",
              325: "\n  ERROR 325: incompatible data found in runoff interface file.",
              327: "\n  ERROR 327: attempting to read beyond end of runoff interface file.",
              329: "\n  ERROR 329: error in reading from runoff interface file.",

              330: "\n  ERROR 330: hotstart interface files have same names.",
              331: "\n  ERROR 331: cannot open hotstart interface file %s.",
              333: "\n  ERROR 333: incompatible data found in hotstart interface file.",
              335: "\n  ERROR 335: error in reading from hotstart interface file.",

              336: "\n  ERROR 336: no climate file specified for evaporation and/or wind speed.",
              337: "\n  ERROR 337: cannot open climate file %s.",
              338: "\n  ERROR 338: error in reading from climate file %s.",
              339: "\n  ERROR 339: attempt to read beyond end of climate file %s.",

              341: "\n  ERROR 341: cannot open scratch RDII interface file.",
              343: "\n  ERROR 343: cannot open RDII interface file %s.",
              345: "\n  ERROR 345: invalid format for RDII interface file.",

              351: "\n  ERROR 351: cannot open routing interface file %s.",
              353: "\n  ERROR 353: invalid format for routing interface file %s.",
              355: "\n  ERROR 355: mis-matched names in routing interface file %s.",
              357: "\n  ERROR 357: inflows and outflows interface files have same name.",

              361: "\n  ERROR 361: could not open external file used for Time Series %s.",
              363: "\n  ERROR 363: invalid data in external file used for Time Series %s.",

              401: "\n  ERROR 401: general system error.",
              402: "\n  ERROR 402: cannot open new project while current project still open.",
              403: "\n  ERROR 403: project not open or last run not ended.",
              405: "\n  ERROR 405: amount of output produced will exceed maximum file size;"
                   "\n             either reduce Ending Date or increase Reporting Time Step.",

              # API Error Keys
              501: "\n API Key Error: Object Type Outside Bonds",
              502: "\n API Key Error: Network Not Initialized (Input file open?)",
              503: "\n API Key Error: Simulation Not Running",
              504: "\n API Key Error: Incorrect object type for parameter chosen",
              505: "\n API Key Error: Object index out of Bounds.",
              506: "\n API Key Error: Invalid Pollutant Index",
              507: "\n API Key Error: Invalid Inflow Type",
              508: "\n API Key Error: Invalid Timeseries Index",
              509: "\n API Key Error: Invalid Pattern Index",
              }



import os

from ..inp import SwmmInput


def split_inp_to_files(inp_fn, **kwargs):
    """
    spit an inp-file into the sections and write per section one file

    creates a subdirectory in the directory of the input file with the name of the input file (without ``.inp``)
    and creates for each section a ``.txt``-file

    Args:
        inp_fn (str): path to inp-file
        **kwargs: keyword arguments of the :func:`~swmm_api.input_file.inp_reader.read_inp_file`-function

    Keyword Args:
        custom_converter (dict): dictionary of {section: converter/section_type} Default: :const:`SECTION_TYPES`
    """
    parent = inp_fn.replace('.inp', '')
    os.mkdir(parent)
    inp = SwmmInput.read_file(inp_fn, **kwargs)
    for s in inp.keys():
        with open(os.path.join(parent, s + '.txt'), 'w') as f:
            f.write(inp._data[s])


def read_split_inp_file(inp_fn):
    """
    use this function to read an split inp-file after splitting the file with the :func:`~split_inp_to_files`-function

    Args:
        inp_fn (str): path of the directory of the split inp-file

    Returns:
        SwmmInput: inp-file data
    """
    inp = SwmmInput()
    for header_file in os.listdir(inp_fn):
        header = header_file.replace('.txt', '')
        with open(os.path.join(inp_fn, header_file), 'r') as f:
            inp[header] = f.read()
    return inp



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd
from pandas import to_datetime
from pandas._libs.tslibs.timedeltas import Timedelta

from .helpers import _get_title_of_part, _remove_lines, _part_to_frame, _continuity_part_to_dict, ReportUnitConversion, \
    _routing_part_to_dict


class SwmmReport:
    """
    SWMM Report file (xxx.rpt).
    """
    def __init__(self, filename):
        """
        create Report instance to read an .rpt-file

        Args:
            filename (str): path to .rpt file

        Notes:
            For more information see SWMM 5.1 User Manual | 9.1 Viewing a Status Report | S. 136

        .. Important::
            The summary results displayed in these tables are based on results found at every
            computational time step and not just on the results from each reporting time step.
        """
        self._filename = filename
        # ________________
        self._raw_parts = dict()
        self._converted_parts = dict()
        # ________________
        # split report file to a dict
        self._report_to_dict()

        # ________________
        self._version_title = None
        self._note = None

        self._routing_time_step_summary = None

        # input summaries
        self._element_count = None
        self._rainfall_file_summary = None

        # ________________
        self._raingage_summary = None
        self._subcatchment_summary = None
        self._node_summary = None
        self._link_summary = None
        self._cross_section_summary = None
        self._transect_summary = None

        self._runoff_quantity_continuity = None
        self._flow_routing_continuity = None
        self._groundwater_continuity = None
        self._quality_routing_continuity = None
        self._runoff_quality_continuity = None

        self._highest_continuity_errors = None
        self._time_step_critical_elements = None
        self._highest_flow_instability_indexes = None

        self._analysis_options = None

        self._node_depth_summary = None
        self._node_inflow_summary = None
        self._node_surcharge_summary = None
        self._node_flooding_summary = None
        self._storage_volume_summary = None
        self._outfall_loading_summary = None

        self._link_flow_summary = None
        self._flow_classification_summary = None
        self._conduit_surcharge_summary = None

        self._subcatchment_runoff_summary = None

        self._groundwater_summary = None

        self._pollutant_summary = None
        self._landuse_summary = None
        self._link_pollutant_load_summary = None
        self._subcatchment_washoff_summary = None

        self._pumping_summary = None

        self._lid_control_summary = None
        self._lid_performance_summary = None

        self._control_actions_taken = None

    def __repr__(self):
        return f'SwmmReport(file="{self._filename}")'

    def _report_to_dict(self):
        """
        convert the report file into a dictionary depending on the different parts
        
        Returns:
            dict[str, str]: dictionary of parts of the report file
        """
        with open(self._filename, 'r') as file:
            lines = file.readlines()

        self._raw_parts['Simulation Infos'] = ''.join(lines[-3:])
        lines = lines[:-3]
        parts0 = ''.join(lines).replace('\n\n  ****', '\n  \n  ****').split('\n  \n  ****')

        for i, part in enumerate(parts0):
            if part.startswith('*'):
                part = '  ****' + part

            self._raw_parts[_get_title_of_part(part, i)] = _remove_lines(part, title=False, empty=True, sep=False)

    def _get_converted_part(self, key):
        if key not in self._converted_parts:
            if key not in self._raw_parts:
                return ''
            self._converted_parts[key] = _remove_lines(self._raw_parts[key], title=True, empty=False)

        return self._converted_parts[key]

    @property
    def available_parts(self):
        """
        Available parts in the report file.

        Returns:
            list[str]: list of parts
        """
        return list(self._raw_parts.keys())

    @property
    def analysis_options(self):
        """
        Analysis Options

        Returns:
            dict[str, str | dict[str, str]]: Analysis Options
        """
        if self._analysis_options is None:
            p = self._get_converted_part('Analysis Options')

            res = dict()
            last_key = None
            last_initial_spaces = 0

            for line in p.split('\n'):
                initial_spaces = len(line) - len(line.lstrip())

                if '..' in line:
                    key = line[:line.find('..')].strip()
                    value = line[line.rfind('..') + 2:].strip()

                    if last_initial_spaces > initial_spaces:
                        last_key = None

                    if last_key is not None:
                        res[last_key].update({key: value})
                    else:
                        res[key] = value

                    last_initial_spaces = initial_spaces

                else:
                    last_key = line.replace(':', '').strip()
                    res[last_key] = dict()

            self._analysis_options = res
        return self._analysis_options

    @property
    def flow_unit(self):
        """
        Flow unit of the simulation.

        Returns:
            str: flow unit
        """
        return self.analysis_options['Flow Units']

    @property
    def unit(self):
        """
        Unit label for some part of the report file.

        Some columns or keys in the report use units. To write general functions for all units use this helper.

        Returns:
            swmm_api.report_file.helpers.ReportUnitConversion: helper for unit labels
        """
        return ReportUnitConversion(self.flow_unit)

    @property
    def element_count(self):
        """
        Element Count

        Returns:
            dict[str, str]: element count
        """
        if self._element_count is None:
            p = self._get_converted_part('Element Count')

            res = dict()
            last_key = None
            last_initial_spaces = 0

            for line in p.split('\n'):
                initial_spaces = len(line) - len(line.lstrip())

                if '..' in line:
                    key = line[:line.find('..')].strip()
                    value = line[line.rfind('..') + 2:].strip()

                    if last_initial_spaces > initial_spaces:
                        last_key = None

                    if last_key is not None:
                        res[last_key].update({key: value})
                    else:
                        res[key] = value

                    last_initial_spaces = initial_spaces

                else:
                    last_key = line.replace(':', '').strip()
                    res[last_key] = dict()

            self._element_count = res
        return self._element_count

    @property
    def flow_routing_continuity(self):
        """
        Flow Routing Continuity

        Returns:
            dict[str, str]: Flow Routing Continuity
        """
        if self._flow_routing_continuity is None:
            raw = self._raw_parts.get('Flow Routing Continuity', None)
            self._flow_routing_continuity = _continuity_part_to_dict(raw)
        return self._flow_routing_continuity

    @property
    def runoff_quantity_continuity(self):
        """
        Runoff Quantity Continuity

        Returns:
            dict[str, str]: Runoff Quantity Continuity
        """
        if self._runoff_quantity_continuity is None:
            raw = self._raw_parts.get('Runoff Quantity Continuity', None)
            self._runoff_quantity_continuity = _continuity_part_to_dict(raw)
        return self._runoff_quantity_continuity

    @property
    def groundwater_continuity(self):
        """
        Groundwater Continuity

        Returns:
            dict[str, str]: Groundwater Continuity
        """
        if self._groundwater_continuity is None:
            p = self._raw_parts.get('Groundwater Continuity', None)
            self._groundwater_continuity = _continuity_part_to_dict(p)
        return self._groundwater_continuity

    @property
    def quality_routing_continuity(self):
        """
        Quality Routing Continuity

        Returns:
            dict[str, str]: Quality Routing Continuity
        """
        if self._quality_routing_continuity is None:
            p = self._raw_parts.get('Quality Routing Continuity', None)
            self._quality_routing_continuity = _continuity_part_to_dict(p)
        return self._quality_routing_continuity

    @property
    def runoff_quality_continuity(self):
        """
        Runoff Quality Continuity

        Returns:
            dict[str, str]: Runoff Quality Continuity
        """
        if self._runoff_quality_continuity is None:
            p = self._raw_parts.get('Runoff Quality Continuity', None)
            self._runoff_quality_continuity = _continuity_part_to_dict(p)
        return self._runoff_quality_continuity

    @property
    def highest_continuity_errors(self):
        """
        Highest Continuity Errors

        Returns:
            dict[str, str]: Highest Continuity Errors
        """
        if self._highest_continuity_errors is None:
            p = self._get_converted_part('Highest Continuity Errors')
            self._highest_continuity_errors = _routing_part_to_dict(p)
        return self._highest_continuity_errors

    @property
    def time_step_critical_elements(self):
        """
        Time-Step Critical Elements

        Returns:
            dict: Time-Step Critical Elements
        """
        if self._time_step_critical_elements is None:
            p = self._get_converted_part('Time-Step Critical Elements')
            self._time_step_critical_elements = _routing_part_to_dict(p)
        return self._time_step_critical_elements

    @property
    def highest_flow_instability_indexes(self):
        """
        Highest Flow Instability Indexes

        Returns:
            dict: Highest Flow Instability Indexes
        """
        if self._highest_flow_instability_indexes is None:
            p = self._get_converted_part('Highest Flow Instability Indexes')
            self._highest_flow_instability_indexes = _routing_part_to_dict(p)
        return self._highest_flow_instability_indexes

    @property
    def node_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._node_summary is None:
            p = self._get_converted_part('Node Summary')
            # p = '-'*10 + '\n' + p
            self._node_summary = _part_to_frame(p)
        return self._node_summary

    @property
    def link_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._link_summary is None:
            p = self._get_converted_part('Link Summary')
            # p = '-'*10 + '\n' + p
            p = p.replace('From Node', 'FromNode')
            p = p.replace('To Node', 'ToNode')
            self._link_summary = _part_to_frame(p)
        return self._link_summary

    @property
    def rainfall_file_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._rainfall_file_summary is None:
            p = self._get_converted_part('Rainfall File Summary')
            # p = '-'*10 + '\n' + p
            # p = p.replace('Data Source', 'DataSource')
            self._rainfall_file_summary = _part_to_frame(p)
        return self._rainfall_file_summary

    @property
    def raingage_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._raingage_summary is None:
            p = self._get_converted_part('Raingage Summary')
            # p = '-'*10 + '\n' + p
            p = p.replace('Data Source', 'DataSource')
            self._raingage_summary = _part_to_frame(p)
        return self._raingage_summary

    @property
    def subcatchment_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._subcatchment_summary is None:
            p = self._get_converted_part('Subcatchment Summary')
            # p = '-'*10 + '\n' + p
            p = p.replace('Rain Gage', 'RainGage')
            self._subcatchment_summary = _part_to_frame(p)
        return self._subcatchment_summary

    @property
    def cross_section_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._cross_section_summary is None:
            p = self._get_converted_part('Cross Section Summary')
            # p = '-'*10 + '\n' + p
            self._cross_section_summary = _part_to_frame(p)
        return self._cross_section_summary

    @property
    def subcatchment_runoff_summary(self):
        """
        Subcatchment Runoff Summary

        Returns:
            pandas.DataFrame: Subcatchment Runoff Summary
        """
        if self._subcatchment_runoff_summary is None:
            p = self._get_converted_part('Subcatchment Runoff Summary')
            self._subcatchment_runoff_summary = _part_to_frame(p)
        return self._subcatchment_runoff_summary

    @property
    def node_depth_summary(self):
        """
        Node Depth Summary

        Returns:
            pandas.DataFrame: Node Depth Summary
        """
        if self._node_depth_summary is None:
            p = self._get_converted_part('Node Depth Summary')
            self._node_depth_summary = _part_to_frame(p)
        return self._node_depth_summary

    @property
    def node_inflow_summary(self):
        """
        Node Inflow Summary

        Returns:
            pandas.DataFrame: Node Inflow Summary
        """
        if self._node_inflow_summary is None:
            p = self._get_converted_part('Node Inflow Summary')
            self._node_inflow_summary = _part_to_frame(p)
        return self._node_inflow_summary

    @property
    def node_surcharge_summary(self):
        """
        Node Surcharge Summary

        Returns:
            pandas.DataFrame: Node Surcharge Summary
        """
        if self._node_surcharge_summary is None:
            p = self._get_converted_part('Node Surcharge Summary')
            # if 'No nodes were surcharged.' in p:
            #     self._node_surcharge_summary = pd.DataFrame()
            # else:
            self._node_surcharge_summary = _part_to_frame(p)
        return self._node_surcharge_summary

    @property
    def node_flooding_summary(self):
        """
        Node Flooding Summary

        Returns:
            pandas.DataFrame: Node Flooding Summary
        """
        if self._node_flooding_summary is None:
            p = self._get_converted_part('Node Flooding Summary')
            # if 'No nodes were flooded.' in p:
            #     self._node_flooding_summary = pd.DataFrame()
            # else:
            self._node_flooding_summary = _part_to_frame(p)
        return self._node_flooding_summary

    @property
    def storage_volume_summary(self):
        """
        Storage Volume Summary

        Returns:
            pandas.DataFrame: Storage Volume Summary
        """
        if self._storage_volume_summary is None:
            p = self._get_converted_part('Storage Volume Summary')

            # for reading the table and accepting names shorten than 8 characters
            p = p.replace('Storage Unit', 'Storage_Unit')

            self._storage_volume_summary = _part_to_frame(p)
        return self._storage_volume_summary

    @property
    def outfall_loading_summary(self):
        """
        Outfall Loading Summary

        Returns:
            pandas.DataFrame: Outfall Loading Summary
        """
        if self._outfall_loading_summary is None:
            p = self._get_converted_part('Outfall Loading Summary')
            self._outfall_loading_summary = _part_to_frame(p.replace('Outfall Node', 'Outfall_Node '))
        return self._outfall_loading_summary

    @property
    def link_flow_summary(self):
        """
        Link Flow Summary

        Returns:
            pandas.DataFrame: Link Flow Summary
        """
        if self._link_flow_summary is None:
            p = self._get_converted_part('Link Flow Summary')
            self._link_flow_summary = _part_to_frame(p)
        return self._link_flow_summary

    @property
    def flow_classification_summary(self):
        """
        Flow Classification Summary

        Returns:
            pandas.DataFrame: Flow Classification Summary
        """
        if self._flow_classification_summary is None:
            p = self._get_converted_part('Flow Classification Summary')
            t = '---------- Fraction of Time in Flow Class ----------'
            p = p.replace(t, ' ' * len(t))
            self._flow_classification_summary = _part_to_frame(p)
        return self._flow_classification_summary

    @property
    def conduit_surcharge_summary(self):
        """
        Conduit Surcharge Summary

        Returns:
            pandas.DataFrame: Conduit Surcharge Summary
        """
        if self._conduit_surcharge_summary is None:
            p = self._get_converted_part('Conduit Surcharge Summary')

            # --------------------------------------------
            # if 'No conduits were surcharged.' in p:
            #     self._conduit_surcharge_summary = pd.DataFrame()
            #
            # else:
            p = p.replace('--------- Hours Full -------- ', 'HoursFull Hours Full HoursFull')
            p = p.replace('Both Ends', 'Both_Ends')
            self._conduit_surcharge_summary = _part_to_frame(p)
        return self._conduit_surcharge_summary

    @property
    def control_actions_taken(self):
        """
        Control Actions Taken

        Returns:
            list[str]: Control Actions Taken
        """
        if self._control_actions_taken is None:
            p = self._get_converted_part('Control Actions Taken')
            if p:
                self._control_actions_taken = p.split('\n')
        return self._control_actions_taken

    @property
    def groundwater_summary(self):
        """
        Groundwater Summary

        Returns:
            pandas.DataFrame: Groundwater Summary
        """
        if self._groundwater_summary is None:
            p = self._get_converted_part('Groundwater Summary')
            self._groundwater_summary = _part_to_frame(p)
        return self._groundwater_summary

    @property
    def landuse_summary(self):
        """
        Landuse Summary

        Returns:
            pandas.DataFrame: Landuse Summary
        """
        if self._landuse_summary is None:
            p = self._get_converted_part('Landuse Summary')
            self._landuse_summary = _part_to_frame(p)
        return self._landuse_summary

    @property
    def lid_control_summary(self):
        """
        LID Control Summary

        Returns:
            pandas.DataFrame: LID Control Summary
        """
        if self._lid_control_summary is None:
            p = self._get_converted_part('LID Control Summary')
            self._lid_control_summary = _part_to_frame(p)
        return self._lid_control_summary

    @property
    def lid_performance_summary(self):
        """
        LID Performance Summary

        Returns:
            pandas.DataFrame: LID Performance Summary
        """
        if self._lid_performance_summary is None:
            p = self._get_converted_part('LID Performance Summary')
            self._lid_performance_summary = _part_to_frame(p)
        return self._lid_performance_summary

    @property
    def link_pollutant_load_summary(self):
        """
        Link Pollutant Load Summary

        Returns:
            pandas.DataFrame: Link Pollutant Load Summary
        """
        if self._link_pollutant_load_summary is None:
            p = self._get_converted_part('Link Pollutant Load Summary')
            self._link_pollutant_load_summary = _part_to_frame(p)
        return self._link_pollutant_load_summary

    @property
    def note(self):
        """
        Note

        Returns:
            str: Node
        """
        if self._note is None:
            self._note = ' '.join(self._get_converted_part('Note').strip(' *').split())
        return self._note

    @property
    def pollutant_summary(self):
        """
        Pollutant Summary

        Returns:
            pandas.DataFrame: Pollutant Summary
        """
        if self._pollutant_summary is None:
            p = self._get_converted_part('Pollutant Summary')
            self._pollutant_summary = _part_to_frame(p)
        return self._pollutant_summary

    @property
    def pumping_summary(self):
        """
        Pumping Summary

        Returns:
            pandas.DataFrame: Pumping Summary
        """
        if self._pumping_summary is None:
            p = self._get_converted_part('Pumping Summary')
            self._pumping_summary = _part_to_frame(p)
        return self._pumping_summary

    @property
    def routing_time_step_summary(self):
        """
        Routing Time Step Summary

        Returns:
            dict[str, float]: Routing Time Step Summary
        """
        if self._routing_time_step_summary is None:
            p = self._get_converted_part('Routing Time Step Summary')
            if p:
                self._routing_time_step_summary = dict()
                for line in self._get_converted_part('Routing Time Step Summary').split('\n'):
                    if 'Time Step Frequencies' in line:
                        continue
                    key, value = ' '.join(line.split()).split(' : ')
                    if 'sec' in value:
                        value = pd.Timedelta(value)
                    elif '%' in value:
                        value = value
                    else:
                        value = float(value)
                    self._routing_time_step_summary[key] = value

        return self._routing_time_step_summary

    @property
    def subcatchment_washoff_summary(self):
        """
        Subcatchment Washoff Summary

        Returns:
            pandas.DataFrame: Subcatchment Washoff Summary
        """
        if self._subcatchment_washoff_summary is None:
            p = self._get_converted_part('Subcatchment Washoff Summary')
            self._subcatchment_washoff_summary = _part_to_frame(p)
        return self._subcatchment_washoff_summary

    @property
    def transect_summary(self):
        """
        Transect Summary

        Returns:
            dict[pandas.DataFrame]: Transect Summary
        """
        if self._transect_summary is None:
            p = self._get_converted_part('Transect Summary')
            self._transect_summary = dict()
            for transect in p.split('Transect')[1:]:
                label, *data = transect.split()
                self._transect_summary[label] = dict()
                sub = data[0][:-1]
                d = []
                for i in data[1:]:
                    if i.endswith(':'):
                        self._transect_summary[label][sub] = d
                        sub = i[:-1]
                    else:
                        d.append(float(i))
                self._transect_summary[label][sub] = d

                self._transect_summary[label] = pd.DataFrame.from_dict(self._transect_summary[label])
        return self._transect_summary

    def get_simulation_info(self):
        """
        Simulation Infos

        Returns:
            dict[str, str]: Simulation Infos
        """
        t = self._raw_parts.get('Simulation Infos', None)
        if t:
            return dict(line.strip().split(':', 1) for line in t.split('\n'))

    @property
    def analyse_start(self):
        """
        Timestamp of the start of the simulation

        Returns:
            pd.Timestamp: Timestamp of the start of the simulation
        """
        v = self.get_simulation_info()['Analysis begun on']
        if v:
            return to_datetime(v)

    @property
    def analyse_end(self):
        """
        Timestamp of the end of the simulation

        Returns:
            pd.Timestamp: Timestamp of the end of the simulation
        """
        v = self.get_simulation_info()['Analysis ended on']
        if v:
            return to_datetime(v)

    @property
    def analyse_duration(self):
        """
        Total elapsed simulation duration.

        Returns:
            pandas.Timedelta: simulation duration
        """
        v = self.get_simulation_info()['Total elapsed time']
        if v:
            if '< 1 sec' in v:
                return Timedelta(seconds=1)

            return Timedelta(v)

    @staticmethod
    def _pprint(di):
        if not di:
            print('{}')
            return
        f = ''
        max_len = len(max(di.keys(), key=len)) + 5
        for key, value in di.items():
            if isinstance(value, (list, tuple, set)):
                key += f' ({len(value)})'
            key += ':'
            if isinstance(value, list) and len(value) > 20:
                start = 0
                for end in range(20, len(value), 20):
                    f += f'{key:<{max_len}}{", ".join(value[start:end])}\",\n'
                    key = ''
                    start = end
            else:
                f += f'{key:<{max_len}}{value}\n'
        print(f)

    def get_errors(self):
        """
        malicious objects per error.

        Returns:
            dict[str, (str | bool | list[str])]: key is the error and value is a object label, a list of object-label or a bool
        """
        t = self._raw_parts.get('Version+Title', None)
        di = dict()
        if t:
            for line in t.split('\n'):
                line = line.strip()
                if line.startswith('ERROR'):
                    label, txt = line.split(':', 1)
                    if label in di:
                        di[label].append(txt)
                    else:
                        di[label] = [txt]
        return di

    def print_errors(self):
        """
        Print the errors in the report file in a pretty way
        """
        self._pprint(self.get_errors())

    def get_version_title(self):
        """
        part of the version and title.

        Returns:
            str: version and title
        """
        if self._version_title is None:
            t = self._raw_parts.get('Version+Title', None)
            self._version_title = t.split('\n')[0].strip()
        return self._version_title

    def get_warnings(self):
        """
        malicious objects per warning.

        Returns:
            dict[str, (str | bool | list[str])]: key is the warning and value is a object label, a list of object-label or a bool

        Notes:
            WARNING 01: wet weather time step reduced to recording interval for Rain Gage xxx.
                The wet weather time step was automatically reduced so that no period with rainfall would be skipped
                during a simulation.
            WARNING 02: maximum depth increased for Node xxx.
                The maximum depth for the node was automatically increased to match the top of the highest connecting
                conduit.
            WARNING 03: negative offset ignored for Link xxx.
                The link’s stipulated offset was below the connecting node's invert so its actual offset was set to 0.
            WARNING 04: minimum elevation drop used for Conduit xxx.
                The elevation drop between the end nodes of the conduit was below 0.001 ft (0.00035 m) so the latter
                value was used instead to calculate its slope.
            WARNING 05: minimum slope used for Conduit xxx.
                The conduit's computed slope was below the user-specified Minimum Conduit Slope so the latter value was
                used instead.
            WARNING 06: dry weather time step increased to wet weather time step.
                The user-specified time step for computing runoff during dry weather periods was lower than that set for
                wet weather periods and was automatically increased to the wet weather value.
            WARNING 07: routing time step reduced to wet weather time step.
                The user-specified time step for flow routing was larger than the wet weather runoff time step and was
                automatically reduced to the runoff time step to prevent loss of accuracy.
            WARNING 08: elevation drop exceeds length for Conduit xxx.
                The elevation drop across the ends of a conduit exceeds its length. The program computes the conduit's
                slope as the elevation drop divided by the length instead of using the more accurate right triangle
                method. The user should check for errors in the length and in both the invert elevations and offsets at
                the conduit's upstream and downstream nodes.
            WARNING 09: time series interval greater than recording interval for Rain Gage xxx.
                The smallest time interval between entries in the precipitation time series used by the rain gage is
                greater than the recording time interval specified for the gage. If this was not actually intended then
                what appear to be continuous periods of rainfall in the time series will instead be read with time gaps
                in between them.
            WARNING 10: crest elevation is below downstream invert for regulator Link xxx.
                The height of the opening on an orifice, weir, or outlet is below the invert elevation of its downstream
                node. Users should check to see if the regulator's offset height or the downstream node's invert
                elevation is in error.
            WARNING 11: non-matching attributes in Control Rule xxx.
                The premise of a control is comparing two different types of attributes to one another (for example,
                conduit flow and junction water depth).
        """
        t = self._raw_parts.get('Version+Title', None)
        di = dict()
        if t:
            for line in t.split('\n'):
                line = line.strip()
                if line.startswith('WARNING'):

                    if ('WARNING 06' in line) or ('WARNING 07' in line):
                        di[line] = True
                    else:
                        *message, object_label = line.split()
                        message = ' '.join(message)
                        if message in di:
                            di[message].append(object_label)
                        else:
                            di[message] = [object_label]
        return di

    def print_warnings(self):
        """
        Print the warnings in the report file in a pretty way.
        """
        self._pprint(self.get_warnings())


def read_rpt_file(filename):
    """
    Read the SWMM Report file (xxx.rpt).

    Args:
        filename (str): filename of the report file

    Returns:
        SwmmReport: report file object
    """
    return SwmmReport(filename)



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"

import pandas as pd
from .following_values import remove_following_zeros


def write_swmm_timeseries_data(series, filename, drop_zeros=True):
    """
    external files in swmm ie. timeseries. (often with the `.dat`-extension)

    Args:
        series (pandas.Series): time-series-data
        filename (str): path and filename for the new file
        drop_zeros (bool): remove all 0 (zero, null) entries in timeseries (SWMM will understand for precipitation)
    """
    if drop_zeros:
        ts = remove_following_zeros(series).dropna()
    else:
        ts = series.dropna()

    if not filename.endswith('.dat'):
        filename += '.dat'

    with open(filename, 'w') as file:
        file.write(';;EPA SWMM Time Series Data\n')
        ts.index.name = ';date      time'
        ts.to_csv(file, sep='\t', index=True, header=True, date_format='%m/%d/%Y %H:%M', line_terminator='\n')


def read_swmm_timeseries_data(filename):
    """
    read text-file of exported timeseries from the EPA-SWMM-GUI

    Args:
        filename (str): path and filename of the file to be read

    Returns:
        pandas.Series: time-series-data
    """
    sep = r'\s+'  # space or tab
    df = pd.read_csv(filename, comment=';', header=None, sep=sep, names=['date', 'time', 'values'])
    df.index = pd.to_datetime(df.pop('date') + ' ' + df.pop('time'))
    return df['values'].copy()


def peek_swmm_timeseries_data(filename, indices):
    """
    take a peek in a text-file of exported timeseries from the EPA-SWMM-GUI

    Args:
        filename (str): path and filename of the file to be read
        indices (list[int]): list of indices to return

    Returns:
        pandas.Series: time-series-data
    """
    df = pd.read_csv(filename, comment=';', header=None, sep=r'\s+', names=['date', 'time', 'values'])
    try:
        df = df.iloc[indices]
    except:
        pass
    df.index = pd.to_datetime(df.pop('date') + ' ' + df.pop('time'))
    return df['values'].copy()


def read_swmm_rainfall_file(filename):
    """
    read text-file of exported timeseries from the EPA-SWMM-GUI

    SWMM 5.1 User Manual | 11.3 Rainfall Files | S. 165

    Args:
        filename (str): path and filename of the file to be read

    Returns:
        pandas.Series: time-series-data with a mulitindex (Datetime, Station)
    """
    sep = r'\s+'  # space or tab
    df = pd.read_csv(filename, comment=';', header=None, sep=sep,
                     names=['station', 'year', 'month', 'day', 'hour', 'minute', 'values'],
                     )
    df.index = pd.to_datetime(df[['year', 'month', 'day', 'hour', 'minute']])
    return df.set_index('station', append=True)['values'].copy()


def read_swmm_tsf(filename, sep='\t'):
    """
    read a swmm .tsf-file (TimeSeriesFile).
    An ASCII-file used in PCSWMM to compare simulations results with measured data, or to export simulation results.

    Args:
        filename (str): path and filename of the file to be read
        sep (str): separator of the columns in the file

    Returns:
        pandas.DataFrame: time-series-data-frame
    """
    df = pd.read_csv(filename, comment=';', header=[0, 1, 2], sep=sep, index_col=0)
    # 5/24/2019 12:00:00 AM
    df.index = pd.to_datetime(df.index, format='%m/%d/%Y %I:%M:%S %p')
    df.columns = ['|'.join(c) for c in df.columns]
    return df



from .section_labels import *
from .sections import *

"""objects or section class for a section in the inp-file"""
SECTION_TYPES = {
    TITLE: TitleSection,
    OPTIONS: OptionSection,
    EVAPORATION: EvaporationSection,
    TEMPERATURE: TemperatureSection,
    REPORT: ReportSection,
    FILES: FilesSection,
    BACKDROP: BackdropSection,
    ADJUSTMENTS: AdjustmentsSection,
    # -----
    # GUI data
    COORDINATES: Coordinate,
    VERTICES: Vertices,
    POLYGONS: Polygon,
    SYMBOLS: Symbol,
    MAP: MapSection,
    LABELS: Label,
    # -----
    # custom section objects
    CONDUITS: Conduit,
    ORIFICES: Orifice,
    WEIRS: Weir,
    PUMPS: Pump,
    OUTLETS: Outlet,

    TRANSECTS: Transect,
    XSECTIONS: CrossSection,
    LOSSES: Loss,
    # -----
    JUNCTIONS: Junction,
    OUTFALLS: Outfall,
    STORAGE: Storage,

    DWF: DryWeatherFlow,
    INFLOWS: Inflow,
    RDII: RainfallDependentInfiltrationInflow,
    TREATMENT: Treatment,
    # -----
    SUBCATCHMENTS: SubCatchment,
    SUBAREAS: SubArea,
    INFILTRATION: InfiltrationGreenAmpt,

    LOADINGS: Loading,
    WASHOFF: WashOff,
    BUILDUP: BuildUp,
    COVERAGES: Coverage,
    GWF: GroundwaterFlow,
    GROUNDWATER: Groundwater,
    SNOWPACKS: SnowPack,
    # -----
    RAINGAGES: RainGage,
    PATTERNS: Pattern,
    POLLUTANTS: Pollutant,
    CONTROLS: Control,
    CURVES: Curve,
    TIMESERIES: Timeseries,
    TAGS: Tag,
    HYDROGRAPHS: Hydrograph,
    LANDUSES: LandUse,
    AQUIFERS: Aquifer,
    # -----
    LID_CONTROLS: LIDControl,
    LID_USAGE: LIDUsage,
}
"""objects or section class for a section in the inp-file"""



import numpy as np

from swmm_api.input_file.section_labels import *


def append_na(x, every=3):
    every_index = every - 1
    return np.reshape(np.insert(np.reshape(x, (-1, every_index)), every_index, None, axis=1), (1, -1))[0]


def plotly_map(inp, fn):  # TODO
    coords = inp[COORDINATES]
    from mp.libs.timeseries.plots.plotly_interface import PlotlyAxes, Ax
    from plotly.graph_objs import Scatter
    axes = PlotlyAxes()

    for edge_kind, color in [('CONDUITS', 'black'),
                             ('WEIRS', 'red'),
                             ('PUMPS', 'blue'),
                             ('ORIFICES', 'orange'),
                             ('OUTLETS', 'cyan')]:
        if edge_kind not in inp:
            continue

        nodes = inp[edge_kind][['FromNode', 'ToNode']].stack().values
        x = coords.lookup(nodes, ['x'] * len(nodes))
        y = coords.lookup(nodes, ['y'] * len(nodes))

        x = append_na(x, 3)
        y = append_na(y, 3)

        edge_scatter = Scatter(x=x, y=y, hoverinfo='none',  # hovertext=[],
                               showlegend=True, line=dict(color=color),
                               mode='lines', name=edge_kind)

        axes.append(Ax(edge_scatter))

    axes.append(Ax(Scatter(x=coords.x, y=coords.y, hoverinfo='text', hovertext=coords.index,
                           mode='markers',
                           name='JUNCTIONS', marker=dict(color='grey'))))

    # https://plot.ly/python/igraph-networkx-comparison/
    fig = axes.get_figure(vertical_spacing=0.02, spaces=[4, 4, 1.5], bg=None)
    # fig.set_title(inp.station)

    fig.turn_axes_off()
    fig.set_size_auto()
    fig.set_hover_mode("closest")
    fig.save(fn, auto_open=True)


def plotly_mapX(inp, fn):
    from mp.libs.timeseries.plots.plotly_interface import PlotlyAxes, Ax
    from plotly.graph_objs import Scattergeo

    # TODO keine ahnung wie man koordinaten umrechnet
    coords = inp[COORDINATES].frame
    axes = PlotlyAxes()

    # for edge_kind, color in [('CONDUITS', 'black'),
    #                          ('WEIRS', 'red'),
    #                          ('PUMPS', 'blue'),
    #                          ('ORIFICES', 'orange'),
    #                          ('OUTLETS', 'cyan')]:
    #     if edge_kind not in inp:
    #         continue
    #
    #     nodes = inp[edge_kind][['FromNode', 'ToNode']].stack().values
    #     x = coords.lookup(nodes, ['x'] * len(nodes))
    #     y = coords.lookup(nodes, ['y'] * len(nodes))
    #
    #     x = append_na(x, 3)
    #     y = append_na(y, 3)
    #
    #     edge_scatter = Scatter(x=x, y=y, hoverinfo='none',  # hovertext=[],
    #                            showlegend=True, line=dict(color=color),
    #                            mode='lines', name=edge_kind)
    #
    #     axes.append(Ax(edge_scatter))

    axes.append(
        Ax(Scattergeo(lon=coords.x, lat=coords.y, hoverinfo='lon+lat',  # 'text', hovertext=coords.index,
                      mode='markers',
                      name='JUNCTIONS', marker=dict(color='grey'))))

    # https://plot.ly/python/igraph-networkx-comparison/
    fig = axes.get_figure(bg=None)
    fig.figure.layout.update(dict(geo=dict(
        scope="europe",
        projection=dict(type="transverse mercator"),
        showland=True,
        lonaxis=dict(
            showgrid=True,
            gridwidth=0.5,
            # range=[-140.0, -55.0],
            dtick=5
        ),
        lataxis=dict(
            showgrid=True,
            gridwidth=0.5,
            # range=[20.0, 60.0],
            dtick=5
        )
        # rotation=dict(
        #     lon=-34/2
        # )
        # landcolor='rgb(243, 243, 243)',
        # countrycolor='rgb(204, 204, 204)',
    ), ))
    # fig.set_title(inp.station)
    # fig.turn_axes_off()
    # fig.set_size_auto()
    # fig.set_hover_mode("closest")
    fig.save(fn, auto_open=True)



from numpy import NaN

from ._identifiers import IDENTIFIERS
from ..helpers import BaseSectionObject
from .._type_converter import to_bool, infer_type
from ..section_labels import JUNCTIONS, OUTFALLS, STORAGE


# NEU in python 3.7
# from dataclasses import dataclass
# @dataclass
# class Junction2(BaseSectionObject):
#     _identifier = IDENTIFIERS.Name
#     Name: str
#     Elevation: float
#     MaxDepth: float = 0
#     InitDepth: float = 0
#     SurDepth: float = 0
#     Aponded: float = 0


class _Node(BaseSectionObject):
    _identifier = IDENTIFIERS.Name

    def __init__(self, Name, Elevation):
        self.Name = str(Name)
        self.Elevation = float(Elevation)


class Junction(_Node):
    """
    Section: [**JUNCTIONS**]

    Purpose:
        Identifies each junction node of the drainage system.
        Junctions are points in space where channels and pipes connect together.
        For sewer systems they can be either connection fittings or manholes.

    Format:
        ::

            Name Elev (Ymax Y0 Ysur Apond)

    Format-PCSWMM:
        ``Name  InvertElev  MaxDepth  InitDepth  SurchargeDepth  PondedArea``

    Format-SWMM-GUI:
        ``Name  Elevation  MaxDepth  InitDepth  SurDepth  Aponded``

    Remarks:
        If Ymax is 0 then SWMM sets the maximum depth equal to the distance
        from the invert to the top of the highest connecting link.

        If the junction is part of a force main section of the system then set Ysur
        to the maximum pressure that the system can sustain.

        Surface ponding can only occur when Apond is non-zero and the ALLOW_PONDING analysis option is turned on.

    Args:
        Name (str): name assigned to junction node.
        Elevation (float): elevation of junction invert (ft or m). ``Elev``
        MaxDepth (float): depth from ground to invert elevation (ft or m) (default is 0). ``Ymax``
        InitDepth (float): water depth at start of simulation (ft or m) (default is 0). ``Y0``
        SurDepth (float): maximum additional head above ground elevation that manhole junction
                            can sustain under surcharge conditions (ft or m) (default is 0). ``Ysur``
        Aponded (float): area subjected to surface ponding once water depth exceeds Ymax (ft2 or m2) (default is 0). ``Apond``

    Attributes:
        Name (str): name assigned to junction node.
        Elevation (float): elevation of junction invert (ft or m). ``Elev``
        MaxDepth (float): depth from ground to invert elevation (ft or m) (default is 0). ``Ymax``
        InitDepth (float): water depth at start of simulation (ft or m) (default is 0). ``Y0``
        SurDepth (float): maximum additional head above ground elevation that manhole junction
                            can sustain under surcharge conditions (ft or m) (default is 0). ``Ysur``
        Aponded (float): area subjected to surface ponding once water depth exceeds Ymax (ft2 or m2) (default is 0). ``Apond``
    """
    _section_label = JUNCTIONS

    def __init__(self, Name, Elevation, MaxDepth=0, InitDepth=0, SurDepth=0, Aponded=0):
        _Node.__init__(self, Name, Elevation)
        self.MaxDepth = float(MaxDepth)
        self.InitDepth = float(InitDepth)
        self.SurDepth = float(SurDepth)
        self.Aponded = float(Aponded)


class Outfall(_Node):
    """
    Section: [**OUTFALLS**]

    Purpose:
        Identifies each outfall node (i.e., final downstream boundary) of the drainage system and the corresponding
        water stage elevation. Only one link can be incident on an outfall node.

    Formats:
        ::

            Name Elev FREE               (Gated) (RouteTo)
            Name Elev NORMAL             (Gated) (RouteTo)
            Name Elev FIXED      Stage   (Gated) (RouteTo)
            Name Elev TIDAL      Tcurve  (Gated) (RouteTo)
            Name Elev TIMESERIES Tseries (Gated) (RouteTo)

    Formats-PCSWMM:
        ``Name  InvertElev  OutfallType  Stage/Table/TimeSeries  TideGate  RouteTo``

    Format-SWMM-GUI:
        ``Name  Elevation  Type  StageData  Gated  RouteTo``

    Args:
        Name (str): name assigned to outfall node.
        Elevation (float): invert elevation (ft or m). ``Elev``
        Type (str): one of ``FREE``, ``NORMAL``, ``FIXED``, ``TIDAL``, ``TIMESERIES``
        *args: -Arguments below-
        Data (float | str): one of the following

            - Stage (float): elevation of fixed stage outfall (ft or m). for ``FIXED``-Type
            - Tcurve (str): name of curve in [``CURVES``] section containing tidal height (i.e., outfall stage) v. hour of day over a complete tidal cycle. for ``TIDAL``-Type
            - Tseries (str): name of time series in [``TIMESERIES``] section that describes how outfall stage varies with time.  for ``TIMESERIES``-Type

        FlapGate (bool, Optional): ``YES`` or ``NO`` depending on whether a flap gate is present that prevents reverse flow. The default is ``NO``. ``Gated``
        RouteTo (str, Optional): name of a subcatchment that receives the outfall's discharge. The default is not to route the outfall’s discharge.

    Attributes:
        Name (str): name assigned to outfall node.
        Elevation (float): invert elevation (ft or m). ``Elev``
        Type (str): one of ``FREE``, ``NORMAL``, ``FIXED``, ``TIDAL``, ``TIMESERIES``
        Data (float | str): one of the following

            - Stage (float): elevation of fixed stage outfall (ft or m). for ``FIXED``-Type
            - Tcurve (str): name of curve in [``CURVES``] section containing tidal height (i.e., outfall stage) v. hour of day over a complete tidal cycle. for ``TIDAL``-Type
            - Tseries (str): name of time series in [``TIMESERIES``] section that describes how outfall stage varies with time.  for ``TIMESERIES``-Type

        FlapGate (bool, Optional): ``YES`` or ``NO`` depending on whether a flap gate is present that prevents reverse flow. The default is ``NO``. ``Gated``
        RouteTo (str, Optional): name of a subcatchment that receives the outfall's discharge. The default is not to route the outfall’s discharge.
    """
    _section_label = OUTFALLS

    class TYPES:
        FREE = 'FREE'
        NORMAL = 'NORMAL'
        FIXED = 'FIXED'
        TIDAL = 'TIDAL'
        TIMESERIES = 'TIMESERIES'

    def __init__(self, Name, Elevation, Type, *args, Data=NaN, FlapGate=False, RouteTo=NaN):
        _Node.__init__(self, Name, Elevation)
        self.Type = Type
        self.Data = NaN

        if args:
            if Type in [Outfall.TYPES.FIXED,
                        Outfall.TYPES.TIDAL,
                        Outfall.TYPES.TIMESERIES]:
                self._data_init(*args)
            else:
                if len(args) == 3:
                    self._data_init(*args)
                else:
                    self._no_data_init(*args)
        else:
            self.Data = Data
            self.FlapGate = to_bool(FlapGate)
            self.RouteTo = RouteTo

    def _no_data_init(self, Gated=False, RouteTo=NaN):
        """
        if no keyword arguments used

        Args:
            Gated (bool): YES or NO depending on whether a flap gate is present that prevents reverse flow. The
            default is NO.
            RouteTo (str): optional name of a subcatchment that receives the outfall's discharge.
                           The default is not to route the outfall’s discharge.
        """
        self.FlapGate = to_bool(Gated)
        self.RouteTo = RouteTo

    def _data_init(self, Data=NaN, Gated=False, RouteTo=NaN):
        """
        if not keyword arguments were used

        Args:
            Data (float | str): one of the following
                Stage (float): elevation of fixed stage outfall (ft or m).
                Tcurve (str): name of curve in [CURVES] section containing tidal height (i.e., outfall stage) v.
                    hour of day over a complete tidal cycle.
                Tseries (str): name of time series in [TIMESERIES] section that describes how outfall stage varies
                with time.
            Gated (bool): YES or NO depending on whether a flap gate is present that prevents reverse flow. The
            default is NO.
            RouteTo (str): optional name of a subcatchment that receives the outfall's discharge.
                           The default is not to route the outfall’s discharge.
        """
        self.Data = Data
        self.FlapGate = to_bool(Gated)
        self.RouteTo = RouteTo


class Storage(_Node):
    """
    Section: [**STORAGE**]

    Purpose:
        Identifies each storage node of the drainage system.
        Storage nodes can have any shape as specified by a surface area versus water depth relation.

    Format:
        ::

            Name Elev Ymax Y0 TABULAR    Acurve   (Apond Fevap Psi Ksat IMD)
            Name Elev Ymax Y0 FUNCTIONAL A1 A2 A0 (Apond Fevap Psi Ksat IMD)

    Format-PCSWMM:
        ``Name Elev MaxDepth InitDepth Shape CurveName/Params N/A Fevap Psi Ksat IMD``

    Format-SWMM-GUI:
        ``Name  InvertElev  MaxDepth  InitDepth  StorageCurve  CurveParams  EvapFrac  InfiltrationParameters``

    Remarks:
        A1, A2, and A0 are used in the following expression that relates surface area (ft2 or m2) to water depth
        (ft or m) for a storage unit with ``FUNCTIONAL`` geometry:

        Area = A0 + A1 * Depth ^ A2

        For ``TABULAR`` geometry, the surface area curve will be extrapolated outwards to meet the unit's maximum depth
        if need be.

        The parameters Psi, Ksat, and IMD need only be supplied if seepage loss through the soil at the bottom and
        sloped sides of the storage unit should be considered.
        They are the same Green-Ampt infiltration parameters described in the [``INFILTRATION``] section.
        If Ksat is zero then no seepage occurs while if IMD is zero then seepage occurs at a constant rate equal to
        Ksat.
        Otherwise seepage rate will vary with storage depth.

    From C-Code:
        ::

            //  Format of input line is:
            //     nodeID  elev  maxDepth  initDepth  FUNCTIONAL a1 a2 a0 surDepth fEvap (infil) //(5.1.013)
            //     nodeID  elev  maxDepth  initDepth  TABULAR    curveID  surDepth fEvap (infil) //

    Args:
        Name (str): name assigned to storage node.
        Elevation (float): invert elevation (ft or m). ``Elev``
        MaxDepth (float): maximum water depth possible (ft or m). ``Ymax``
        InitDepth (float): water depth at the start of the simulation (ft or m). ``Y0``
        Type (str): ``TABULAR`` | ``FUNCTIONAL``
        *args (): -Arguments below-
        Curve (str | list):

            - :obj:`str`: name of curve in [``CURVES``] section with surface area (ft2 or m2) as a function of depth (ft or m) for ``TABULAR`` geometry. ``Acurve``
            - :obj:`list`: ``FUNCTIONAL`` relation between surface area and depth with

               - A1 (:obj:`float`): coefficient
               - A2 (:obj:`float`): exponent
               - A0 (:obj:`float`): constant

        Apond (float): this parameter has been deprecated – use 0.
        Fevap (float): fraction of potential evaporation from surface realized (default is 0).
        Psi (float): soil suction head (inches or mm).
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr).
        IMD (float): soil initial moisture deficit (fraction).

    Attributes:
        Name (str): name assigned to storage node.
        Elevation (float): invert elevation (ft or m). ``Elev``
        MaxDepth (float): maximum water depth possible (ft or m). ``Ymax``
        InitDepth (float): water depth at the start of the simulation (ft or m). ``Y0``
        Type (str): ``TABULAR`` or ``FUNCTIONAL``
        Curve (str | list):

            - :obj:`str`: name of curve in [``CURVES``] section with surface area (ft2 or m2) as a function of depth (ft or m) for ``TABULAR`` geometry. ``Acurve``
            - :obj:`list`: ``FUNCTIONAL`` relation between surface area and depth with

               - A1 (:obj:`float`): coefficient
               - A2 (:obj:`float`): exponent
               - A0 (:obj:`float`): constant

        Apond (float): this parameter has been deprecated – use 0.
        Fevap (float): fraction of potential evaporation from surface realized (default is 0).
        Psi (float): soil suction head (inches or mm).
        Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr).
        IMD (float): soil initial moisture deficit (fraction).
    """
    _section_label = STORAGE

    class TYPES:
        TABULAR = 'TABULAR'
        FUNCTIONAL = 'FUNCTIONAL'

    def __init__(self, Name: str, Elevation: float, MaxDepth: float, InitDepth: float, Type: str, *args, Curve=None,
                 Apond: float = 0, Fevap: float=0, Psi: float=NaN, Ksat: float=NaN, IMD: float=NaN):
        _Node.__init__(self, Name, Elevation)
        self.MaxDepth = float(MaxDepth)
        self.InitDepth = float(InitDepth)
        self.Type = Type

        if args:
            if Type == Storage.TYPES.TABULAR:
                self._tabular_init(*args)

            elif Type == Storage.TYPES.FUNCTIONAL:
                self._functional_init(*args)

            else:
                raise NotImplementedError()
        else:
            self.Curve = Curve
            self._optional_args(Apond, Fevap, Psi, Ksat, IMD)

    def _functional_init(self, A1, A2, A0, *args, **kwargs):
        """
        for storage type ``'FUNCTIONAL'``

        Args:
            A1 (float): coefficient of FUNCTIONAL relation between surface area and depth.
            A2 (float): exponent of FUNCTIONAL relation between surface area and depth.
            A0 (float): constant of FUNCTIONAL relation between surface area and depth.

            Apond (float): this parameter has been deprecated – use 0.
            Fevap (float): fraction of potential evaporation from surface realized (default is 0).
        """
        self.Curve = infer_type([A1, A2, A0])
        self._optional_args(*args, **kwargs)

    def _tabular_init(self, Curve, *args, **kwargs):
        """
        for storage type ``'TABULAR'``

        Args:
            Curve (str): name of curve in [CURVES] section with surface area (ft2 or m2)
                as a function of depth (ft or m) for TABULAR geometry.
            Apond (float): this parameter has been deprecated – use 0.
            Fevap (float): fraction of potential evaporation from surface realized (default is 0).
        """
        self.Curve = Curve
        self._optional_args(*args, **kwargs)

    def _optional_args(self, Apond=0, Fevap=0, *exfiltration_args, **exfiltration_kwargs):
        """
        for the optional arguemts

        Args:
            Apond (float): this parameter has been deprecated – use 0.
            Fevap (float): fraction of potential evaporation from surface realized (default is 0).
        """
        self.Apond = float(Apond)
        self.Fevap = float(Fevap)
        self._exfiltration_args(*exfiltration_args, **exfiltration_kwargs)

    def _exfiltration_args(self, Psi=NaN, Ksat=NaN, IMD=NaN):
        """
        for the optional arguemts

        Args:
            Psi (float): soil suction head (inches or mm).
            Ksat (float): soil saturated hydraulic conductivity (in/hr or mm/hr).
            IMD (float): soil initial moisture deficit (fraction).
        """
        self.Psi = float(Psi)
        self.Ksat = float(Ksat)
        self.IMD = float(IMD)



from statistics import mean

import pandas as pd

from .collection import nodes_dict, links_dict
from .graph import links_connected
from .tags import get_subcatchment_tags, get_node_tags
from ..inp import SwmmInput
from ..section_labels import *
from ..section_lists import LINK_SECTIONS, NODE_SECTIONS
from ..sections.link import _Link
from ..sections.link_component import CrossSection

"""
a collection of macros to manipulate an inp-file

use this file as an example for the usage of this package
"""


########################################################################################################################
def find_node(inp: SwmmInput, label):
    """
    find node in inp-file data

    Args:
        inp (SwmmInput): inp-file data
        label (str): node Name/label

    Returns:
        _Node | Junction | swmm_api.input_file.sections.node.Storage | Outfall: searched node (if not found None)
    """
    return nodes_dict(inp).get(label, None)


def find_link(inp: SwmmInput, label):
    """
    find link in inp-file data

    Args:
        inp (SwmmInput): inp-file data
        label (str): link Name/label

    Returns:
        _Link | Conduit | Weir | Outlet | Orifice | Pump: searched link (if not found None)
    """
    return links_dict(inp).get(label, None)


########################################################################################################################
def calc_slope(inp: SwmmInput, conduit):
    """
    calculate the slop of a conduit

    Args:
        inp (SwmmInput): inp-file data
        conduit (Conduit): link

    Returns:
        float: slop of the link
    """
    nodes = nodes_dict(inp)
    return (nodes[conduit.FromNode].Elevation + conduit.InOffset - (
            nodes[conduit.ToNode].Elevation + conduit.OutOffset)) / conduit.Length


def conduit_slopes(inp: SwmmInput):
    """
    get the slope of all conduits

    Args:
        inp (SwmmInput):

    Returns:
        pandas.Series: slopes
    """
    slopes = dict()
    for conduit in inp.CONDUITS.values():
        slopes[conduit.Name] = calc_slope(inp, conduit)
    return pd.Series(slopes)


########################################################################################################################
def _rel_diff(a, b):
    m = mean([a + b])
    if m == 0:
        return abs(a - b)
    return abs(a - b) / m


def _rel_slope_diff(inp: SwmmInput, l0, l1):
    nodes = nodes_dict(inp)
    slope_res = (nodes[l0.FromNode].Elevation + l0.InOffset
                 - (nodes[l1.ToNode].Elevation + l1.OutOffset)
                 ) / (l0.Length + l1.Length)
    return _rel_diff(calc_slope(inp, l0), slope_res)


########################################################################################################################
def conduits_are_equal(inp: SwmmInput, link0, link1, diff_roughness=0.1, diff_slope=0.1, diff_height=0.1):
    """
    check if the links (with all there components) are equal

    Args:
        inp (SwmmInput):
        link0 (Conduit | Weir | Outlet | Orifice | Pump | _Link): first link
        link1 (Conduit | Weir | Outlet | Orifice | Pump | _Link): second link
        diff_roughness (float): difference from which it is considered different.
        diff_slope (float): difference from which it is considered different.
        diff_height (float): difference from which it is considered different.

    Returns:
        bool: if the links are equal
    """
    all_checks_out = True

    # Roughness values match within a specified percent tolerance
    if diff_roughness is not None:
        all_checks_out &= _rel_diff(link0.Roughness, link1.Roughness) < diff_roughness

    xs0 = inp[XSECTIONS][link0.Name]  # type: CrossSection
    xs1 = inp[XSECTIONS][link1.Name]  # type: CrossSection

    # Diameter values match within a specified percent tolerance (1 %)
    if diff_height is not None:
        all_checks_out &= _rel_diff(xs0.Geom1, xs1.Geom1) < diff_height

    # Cross-section shapes must match exactly
    all_checks_out &= xs0.Shape == xs1.Shape

    # Shape curves must match exactly
    if xs0.Shape == CrossSection.SHAPES.CUSTOM:
        all_checks_out &= xs0.Curve == xs1.Curve

    # Transects must match exactly
    elif xs0.Shape == CrossSection.SHAPES.IRREGULAR:
        all_checks_out &= xs0.Tsect == xs1.Tsect

    # Slope values match within a specified tolerance
    if diff_slope is not None:
        rel_slope_diff = _rel_diff(calc_slope(inp, link0), calc_slope(inp, link1))

        # if rel_slope_diff < 0:
        #     nodes = nodes_dict(inp)
        #     print(nodes[link0.FromNode].Elevation, link0.InOffset, nodes[link0.ToNode].Elevation, link0.OutOffset)
        #     print(nodes[link1.FromNode].Elevation, link1.InOffset, nodes[link1.ToNode].Elevation, link1.OutOffset)
        #     print('rel_slope_diff < 0', link0, link1)
        all_checks_out &= rel_slope_diff < diff_slope

    return all_checks_out


def update_no_duplicates(inp_base, inp_update) -> SwmmInput:
    inp_new = inp_base.copy()
    inp_new.update(inp_update)

    for node in nodes_dict(inp_new):
        if sum((node in inp_new[s] for s in NODE_SECTIONS + [SUBCATCHMENTS] if s in inp_new)) != 1:
            for s in NODE_SECTIONS + [SUBCATCHMENTS]:
                if (s in inp_new) and (node in inp_new[s]) and (node not in inp_update[s]):
                    del inp_new[s][node]

    for link in links_dict(inp_new):
        if sum((link in inp_new[s] for s in LINK_SECTIONS if s in inp_new)) != 1:
            for s in LINK_SECTIONS:
                if (s in inp_new) and (link in inp_new[s]) and (link not in inp_update[s]):
                    del inp_new[s][link]

    return inp_new


########################################################################################################################
def increase_max_node_depth(inp, node_label):
    # swmm raises maximum node depth to surrounding xsection height
    previous_, next_ = links_connected(inp, node_label)
    node = nodes_dict(inp)[node_label]
    max_height = node.MaxDepth
    for link in previous_ + next_:
        max_height = max((max_height, inp[XSECTIONS][link.Name].Geom1))
    print(f'MaxDepth increased for node "{node_label}" from {node.MaxDepth} to {max_height}')
    node.MaxDepth = max_height


def set_times(inp, start, end, head=None, tail=None):
    """
    set start and end time of the inp-file

    Args:
        inp (SwmmInput): inp data
        start (datetime.datetime): start time of the simulation and the reporting
        end (datetime.datetime): end time of the simulation
        head (datetime.timedelta): brings start time forward
        tail (datetime.timedelta): brings end time backward

    Returns:
        SwmmInput: changed inp data
    """
    if head is None:
        sim_start = start
    else:
        sim_start = start - head

    if tail is None:
        end = end
    else:
        end = end + tail

    report_start = start
    inp[OPTIONS]['START_DATE'] = sim_start.date()
    inp[OPTIONS]['START_TIME'] = sim_start.time()
    inp[OPTIONS]['REPORT_START_DATE'] = report_start.date()
    inp[OPTIONS]['REPORT_START_TIME'] = report_start.time()
    inp[OPTIONS]['END_DATE'] = end.date()
    inp[OPTIONS]['END_TIME'] = end.time()
    return inp


def combined_subcatchment_frame(inp: SwmmInput):
    """
    combine all information of the subcatchment data-frames

    Args:
        inp (SwmmInput): inp-file data

    Returns:
        pandas.DataFrame: combined subcatchment data
    """
    df = inp[SUBCATCHMENTS].frame.join(inp[SUBAREAS].frame).join(inp[INFILTRATION].frame)
    df = df.join(get_subcatchment_tags(inp))
    return df


def combined_nodes_frame(inp: SwmmInput):
    pass  # TODO


def nodes_data_frame(inp, label_sep='.'):
    nodes_tags = get_node_tags(inp)
    res = None
    for s, _ in iter_sections(inp, NODE_SECTIONS):
        df = inp[s].frame.rename(columns=lambda c: f'{label_sep}{c}')

        if s == STORAGE:
            df[f'{STORAGE}{label_sep}Curve'] = df[f'{STORAGE}{label_sep}Curve'].astype(str)

        for sub_sec in [DWF, INFLOWS]:
            if sub_sec in inp:
                x = inp[sub_sec].frame.unstack(1)
                x.columns = [f'{label_sep}'.join([sub_sec, c[1], c[0]]) for c in x.columns]
                df = df.join(x)

        df = df.join(inp[COORDINATES].frame).join(nodes_tags)

        if res is None:
            res = df
        else:
            res = res.append(df)
    return res


def iter_sections(inp, section_list):
    for s in section_list:
        if s in inp:
            yield s, inp[s]


def delete_sections(inp, section_list):
    for s in section_list:
        if s in inp:
            del inp[s]



__author__ = "Markus Pichler"
__credits__ = ["Markus Pichler"]
__maintainer__ = "Markus Pichler"
__email__ = "markus.pichler@tugraz.at"
__version__ = "0.1"
__license__ = "MIT"


from itertools import product
from numpy import dtype, fromfile
from pandas import date_range, DataFrame, MultiIndex
from .extract import SwmmOutExtract
from .definitions import OBJECTS, VARIABLES

from . import parquet


class SwmmOutput(SwmmOutExtract):
    """
    SWMM Output file (xxx.out).

    combined the reader of `swmmtoolbox`_ with the functionality of pandas

    .. _swmmtoolbox: https://github.com/timcera/swmmtoolbox
    """
    def __init__(self, filename):
        """
        SWMM Output file (xxx.out).

        Args:
            filename(str): path to .rpt file

        Notes:
            based on the python package swmmtoolbox
        """
        SwmmOutExtract.__init__(self, filename)

        self._frame = None
        self._data = None

        # the main datetime index for the results
        self.index = date_range(self.start_date + self.report_interval,
                                periods=self.n_periods, freq=self.report_interval)

    def __repr__(self):
        return f'SwmmOut(file="{self.filename}")'

    def _get_dtypes(self):
        """
        get the dtypes of the data

        Returns:
            str: numpy types
        """
        return 'f8' + ',f4' * self.number_columns

    @property
    def number_columns(self):
        """
        get number of columns of the full results table

        Returns:
            int: number of columns of the full results table
        """
        return sum([len(self.variables[kind]) * len(self.labels[kind]) for kind in OBJECTS.LIST_])

    @property
    def _columns_raw(self):
        """
        get the column-names of the data

        Returns:
            list[list[str]]: multi-level column-names
        """
        columns = list()
        for kind in OBJECTS.LIST_:
            columns += list(product([kind], self.labels[kind], self.variables[kind]))
        return columns

    def to_numpy(self):
        """
        read the full binary .out-file of EPA-SWMM and return a numpy array

        Returns:
            numpy.ndarray: all data
        """
        if self._data is None:
            types = [('datetime', 'f8')]
            types += list(map(lambda i: ('/'.join(i), 'f4'), self._columns_raw))
            self.fp.seek(self.pos_start_output, 0)
            self._data = fromfile(self.fp, dtype=dtype(types))
        return self._data

    def to_frame(self):
        """
        convert all the data to a pandas-DataFrame

        Warnings:
            for a big out-file with many objects, this function may take a long time

        Returns:
            pandas.DataFrame: data
        """
        if self._frame is None:
            self._frame = self._to_pandas(self.to_numpy())
            del self._frame['datetime']
        return self._frame

    def get_part(self, kind=None, label=None, variable=None, slim=False):
        """
        get specific columns of the data to a pandas-DataFame (or pandas-Series for a single column)

        use this function instead of "get_part" if there are a lot of objects in the out-file.

        Args:
            kind (str | list): ["subcatchment", "node", "link", "system"] (predefined in :obj:`swmm_api.output_file.definitions.OBJECTS`)
            label (str | list): name of the objekts
            variable (str | list): variable names (predefined in :obj:`swmm_api.output_file.definitions.VARIABLES`)

                * subcatchment:
                    - ``rainfall`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.RAINFALL`
                    - ``snow_depth`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.SNOW_DEPTH`
                    - ``evaporation`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.EVAPORATION`
                    - ``infiltration`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.INFILTRATION`
                    - ``runoff`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.RUNOFF`
                    - ``groundwater_outflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.GW_OUTFLOW`
                    - ``groundwater_elevation`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.GW_ELEVATION`
                    - ``soil_moisture`` i.e.: :attr:`~swmm_api.output_file.definitions.SUBCATCHMENT_VARIABLES.SOIL_MOISTURE`
                * node:
                    - ``depth`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.DEPTH`
                    - ``head`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.HEAD`
                    - ``volume`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.VOLUME`
                    - ``lateral_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.LATERAL_INFLOW`
                    - ``total_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.TOTAL_INFLOW`
                    - ``flooding`` i.e.: :attr:`~swmm_api.output_file.definitions.NODE_VARIABLES.FLOODING`
                * link:
                    - ``flow`` i.e.: :attr:`~swmm_api.output_file.definitions.LINK_VARIABLES.FLOW`
                    - ``depth`` i.e.: :attr:`~swmm_api.output_file.definitions.LINK_VARIABLES.DEPTH`
                    - ``velocity`` i.e.: :attr:`~swmm_api.output_file.definitions.LINK_VARIABLES.VELOCITY`
                    - ``volume`` i.e.: :attr:`~swmm_api.output_file.definitions.LINK_VARIABLES.VOLUME`
                    - ``capacity`` i.e.: :attr:`~swmm_api.output_file.definitions.LINK_VARIABLES.CAPACITY`
                * system:
                    - ``air_temperature`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.AIR_TEMPERATURE`
                    - ``rainfall`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.RAINFALL`
                    - ``snow_depth`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.SNOW_DEPTH`
                    - ``infiltration`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.INFILTRATION`
                    - ``runoff`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.RUNOFF`
                    - ``dry_weather_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.DW_INFLOW`
                    - ``groundwater_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.GW_INFLOW`
                    - ``RDII_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.RDII_INFLOW`
                    - ``direct_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.DIRECT_INFLOW`
                    - ``lateral_inflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.LATERAL_INFLOW`
                    - ``flooding`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.FLOODING`
                    - ``outflow`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.OUTFLOW`
                    - ``volume`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.VOLUME`
                    - ``evaporation`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.EVAPORATION`
                    - ``PET`` i.e.: :attr:`~swmm_api.output_file.definitions.SYSTEM_VARIABLES.PET`

            slim (bool): set to `True` if there are a lot of objects and just few time-steps in the out-file.


        Returns:
            pandas.DataFrame | pandas.Series: filtered data
        """
        columns = self._filter_part_columns(kind, label, variable)
        if slim:
            values = self.get_selective_results(columns)
        else:
            values = self.to_numpy()[list(map('/'.join, columns))]

        return self._to_pandas(values, drop_useless=True)

    def _filter_part_columns(self, kind=None, label=None, variable=None):
        """
        filter which columns should be extracted

        Args:
            kind (str | list): ["subcatchment", "node", "link", "system"]
            label (str | list): name of the objekts
            variable (str | list): variable names

        Returns:
            list: filtered list of tuple(kind, label, variable)
        """
        def _filter(i, possibilities):
            if i is None:
                return possibilities
            elif isinstance(i, str):
                if i in possibilities:
                    return [i]
                else:
                    return []
            elif isinstance(i, list):
                return [j for j in i if j in possibilities]

        columns = list()
        for k in _filter(kind, OBJECTS.LIST_):
            columns += list(product([k], _filter(label, self.labels[k]), _filter(variable, self.variables[k])))
        return columns

    def _to_pandas(self, data, drop_useless=False):
        """
        convert interim results to pandas DataFrame or Series

        Args:
            data (dict, numpy.ndarray): timeseries data of swmm out file
            drop_useless (bool): if single column data should be returned as Series

        Returns:
            (pandas.DataFrame | pandas.Series): pandas Timerseries of data
        """
        if isinstance(data, dict):
            if not bool(data):
                return DataFrame()
            df = DataFrame.from_dict(data).set_index(self.index)
        else:
            if data.size == 0:
                return DataFrame()
            df = DataFrame(data, index=self.index, dtype=float)

        # -----------
        if df.columns.size == 1:
            return df.iloc[:, 0]
        # -----------
        df.columns = MultiIndex.from_tuples([col.split('/') for col in df.columns])
        if drop_useless:
            df.columns = df.columns.droplevel([i for i, l in enumerate(df.columns.levshape) if l == 1])

        return df

    def to_parquet(self):
        """
        read the binary .out file from EPA-SWMM and write the data to a parquet file

        multi-column-names are separated by a slash ("/")
        read parquet files with parquet.read to get the original column-name-structure
        """
        parquet.write(self.to_frame(), self.filename.replace('.out', '.parquet'))


def read_out_file(filename):
    """
    Read the SWMM Output file (xxx.rpt).

    Args:
        filename (str): filename of the output file

    Returns:
        SwmmOutput: output file object
    """
    return SwmmOutput(filename)


def out2frame(filename):
    """
    Get the content of the SWMM Output file as a DataFrame

    Args:
        filename (str): filename of the output file

    Returns:
        pandas.DataFrame: Content of the SWMM Output file

    .. Important::
        don't use this function if many object are in the out file and you only need few of them.
        In this case use the method :meth:`SwmmOutput.get_part` instead.
    """
    out = SwmmOutput(filename)
    return out.to_frame()



from collections import ChainMap

from ..section_labels import SUBCATCHMENTS
from ..section_lists import NODE_SECTIONS, LINK_SECTIONS


def nodes_dict(inp):
    """
    get a dict of all nodes

    the objects are referenced, so you can use it to modify too.

    Args:
        inp (swmm_api.SwmmInput): inp-file data

    Returns:
        ChainMap[str, swmm_api.input_file.sections.node._Node]: dict of {labels: objects}
    """
    nodes = ChainMap()
    for section in NODE_SECTIONS:
        if section in inp:
            nodes.maps.append(inp[section])
    return nodes


def nodes_subcatchments_dict(inp):
    """
    get a dict of all nodes and subcatchments

    the objects are referenced, so you can use it to modify too.

    Args:
        inp (swmm_api.SwmmInput): inp-file data

    Returns:
        ChainMap[str, swmm_api.input_file.sections.node._Node | swmm_api.input_file.sections.subcatch.SubCatchment]: dict of {labels: objects}
    """
    nodes_subcatchments = nodes_dict(inp)
    if SUBCATCHMENTS in inp:
        nodes_subcatchments.maps.append(inp[SUBCATCHMENTS])
    return nodes_subcatchments


def links_dict(inp):
    """
    get a dict of all links

    the objects are referenced, so you can use it to modify too.

    Args:
        inp (swmm_api.SwmmInput): inp-file data

    Returns:
        ChainMap[str, swmm_api.input_file.sections.link._Link]: dict of {labels: objects}
    """
    links = ChainMap()
    for section in LINK_SECTIONS:
        if section in inp:
            links.maps.append(inp[section])
    return links


def subcatchments_per_node_dict(inp):
    """
    get dict where key=node and value=list of subcatchment connected to the node (set as outlet)

    Args:
        inp (swmm_api.SwmmInput): inp data

    Returns:
        dict[str, list[swmm_api.input_file.sections.subcatch.SubCatchment]]: dict[node] = list(subcatchments)
    """
    if SUBCATCHMENTS in inp:
        di = {n: [] for n in nodes_dict(inp)}
        for s in inp.SUBCATCHMENTS.items():
            di[s.Outlet].append(s)
        return di



# from: https://timcera.bitbucket.io/swmmtoolbox/docsrc/index.html
# https://bitbucket.org/timcera/swmmtoolbox/src/master/swmmtoolbox/swmmtoolbox.py
# copied to reduce dependencies
# ORIGINAL Author Tim Cera with BSD License
# Rewritten for custom use

# SWMM Version > 5.10.10
# Python Version >= 3.7

import copy
from os import remove

import datetime
import struct
from io import SEEK_END, SEEK_SET
from tqdm.auto import tqdm
from warnings import warn

from .definitions import OBJECTS, VARIABLES
from .._read_bin import BinaryReader

VARIABLES_DICT = {
    OBJECTS.SUBCATCHMENT: VARIABLES.SUBCATCHMENT.LIST_,
    OBJECTS.NODE        : VARIABLES.NODE.LIST_,
    OBJECTS.LINK        : VARIABLES.LINK.LIST_,
    OBJECTS.POLLUTANT   : [],
    OBJECTS.SYSTEM      : VARIABLES.SYSTEM.LIST_,
}

_RECORDSIZE = 4
_FLOW_UNITS_METRIC = ['CMS', 'LPS', 'MLD']
_FLOW_UNITS_IMPERIAL = ['CFS', 'GPM', 'MGD']
_FLOW_UNITS = _FLOW_UNITS_IMPERIAL + _FLOW_UNITS_METRIC + [None]
_CONCENTRATION_UNITS = ['MG', 'UG', 'COUNTS']
_MAGIC_NUMBER = 516114522
_PROPERTY_LABELS = ['type', 'area', 'invert', 'max_depth', 'offset', 'length']
_NODES_TYPES = ['JUNCTION', 'OUTFALL', 'STORAGE', 'DIVIDER']
_LINK_TYPES = ['CONDUIT', 'PUMP', 'ORIFICE', 'WEIR', 'OUTLET']


class SwmmExtractValueError(Exception):
    def __init__(self, message):
        super().__init__("\n*\n*   {}\n*\n".format(message))


class SwmmOutExtractWarning(UserWarning):
    pass


class SwmmOutExtract(BinaryReader):
    """The class that handles all extraction of data from the out file."""

    def __init__(self, filename):
        super().__init__(filename)

        # ____
        self.fp.seek(-6 * _RECORDSIZE, SEEK_END)
        (
            _pos_start_labels,  # starting file position of ID names
            _pos_start_input,  # starting file position of input data
            _pos_start_output,  # starting file position of output data
            _n_periods,  # Number of reporting periods
            error_code,
            magic_num_end,
        ) = self._next(6)

        # ____
        self.fp.seek(0, SEEK_SET)
        magic_num_start = self._next()

        # ____
        # check errors
        if magic_num_start != _MAGIC_NUMBER:
            raise SwmmExtractValueError('Beginning magic number incorrect.')

        if magic_num_end != _MAGIC_NUMBER:
            warn('Ending magic number incorrect.', SwmmOutExtractWarning)
            # raise SwmmExtractValueError('Ending magic number incorrect.')
            _n_periods = 0
        elif error_code != 0:
            warn(f'Error code "{error_code}" in output file indicates a problem with the run.', SwmmOutExtractWarning)
            # raise SwmmExtractValueError(f'Error code "{error_code}" in output file indicates a problem with the run.')

        # ---
        # read additional parameters from start of file
        # Version number i.e. "51015"
        self.swmm_version, self.flow_unit, n_subcatch, n_nodes, n_links, n_pollutants = self._next(6)
        self.flow_unit = _FLOW_UNITS[self.flow_unit]

        # ____
        # self.fp.seek(_pos_start_labels, SEEK_SET)  # not needed!
        # print(self.fp.tell(), _pos_start_labels)
        # assert _pos_start_labels == self.fp.tell()
        # ____
        # Read in the names
        # get the dictionary of the object labels for each object type (link, node, subcatchment)
        self.labels = dict()
        for kind, n in zip(OBJECTS.LIST_, [n_subcatch, n_nodes, n_links, n_pollutants, 0]):
            self.labels[kind] = [self._next(n=self._next(), dtype='s') for _ in range(n)]

        # ____
        # print(self.fp.tell(), _pos_start_input)
        # assert _pos_start_input == self.fp.tell()
        # ____
        # Update variables to add pollutant names to subcatchment, nodes, and links.
        # get the dictionary of the object variables for each object type (link, node, subcatchment)
        self.variables = copy.deepcopy(VARIABLES_DICT)
        for kind in [OBJECTS.SUBCATCHMENT, OBJECTS.NODE, OBJECTS.LINK]:
            self.variables[kind] += self.labels[OBJECTS.POLLUTANT]

        # ____
        # System vars do not have names per se, but made names = number labels
        self.labels[OBJECTS.SYSTEM] = ['']  # self.variables[OBJECTS.SYSTEM]

        # ____
        # Read codes of pollutant concentration UNITS = Number of pollutants * 4 byte integers
        _pollutant_unit_labels = [_CONCENTRATION_UNITS[p] if p < len(_CONCENTRATION_UNITS) else 'NaN'
                                  for p in self._next(n_pollutants, flat=False)]
        self.pollutant_units = dict(zip(self.labels[OBJECTS.POLLUTANT], _pollutant_unit_labels))

        # ____
        # property values for subcatchments, nodes and links
        #   subcatchment
        #     area
        #   node
        #     type, invert, & max. depth
        #   link
        #     type, offsets [ht. above start node invert (ft), ht. above end node invert (ft)], max. depth, & length
        self.model_properties = dict()
        for kind in [OBJECTS.SUBCATCHMENT, OBJECTS.NODE, OBJECTS.LINK]:
            self.model_properties[kind] = dict()
            # ------
            # read the property labels per object type
            property_labels = list()
            for i in list(self._next(self._next(), flat=False)):
                property_label = _PROPERTY_LABELS[i]
                if property_label in property_labels:
                    property_label += '_2'
                property_labels.append(property_label)
            # ------
            # read the values per object and per property
            for label in self.labels[kind]:
                self.model_properties[kind][label] = dict()
                for property_label in property_labels:
                    value = self._next(dtype={'type': 'i'}.get(property_label, 'f'))
                    if property_label == 'type':
                        value = {OBJECTS.NODE: _NODES_TYPES, OBJECTS.LINK: _LINK_TYPES}[kind][value]
                    self.model_properties[kind][label][property_label] = value

        # ____
        # double check variables
        for kind in [OBJECTS.SUBCATCHMENT, OBJECTS.NODE, OBJECTS.LINK, OBJECTS.SYSTEM]:
            n_vars = self._next()
            assert n_vars == len(self.variables[kind])
            self._next(n_vars)

        # ____
        self.start_date = datetime.datetime(1899, 12, 30) + datetime.timedelta(days=self._next(dtype='d'))
        self.report_interval = datetime.timedelta(seconds=self._next())

        # ____
        self._bytes_per_period = self._infer_bytes_per_period()

        # ____
        # print(self.fp.tell(), _pos_start_output)
        # assert _pos_start_output == self.fp.tell()
        # if _pos_start_output == 0:
        # Out File not complete!
        self.pos_start_output = self.fp.tell()

        self.n_periods = _n_periods
        if _n_periods == 0:
            self._infer_n_periods()
            warn('Infer time periods of the output file due to an corrupt SWMM .out-file.', SwmmOutExtractWarning)

        if self.n_periods == 0:
            warn('There are zero time periods in the output file.', SwmmOutExtractWarning)
            # raise SwmmExtractValueError('There are zero time periods in the output file.')

    def __repr__(self):
        return f'SwmmOutExtract(file="{self.filename}")'

    def _infer_bytes_per_period(self):
        """
        Calculate the bytes for each time period when reading the computed results

        Returns:
            int: bytes per period
        """
        _bytes_per_period = 2  # for the datetime
        for obj in [OBJECTS.SUBCATCHMENT, OBJECTS.NODE, OBJECTS.LINK]:
            _bytes_per_period += len(self.variables[obj]) * len(self.labels[obj])
        _bytes_per_period += len(self.variables[OBJECTS.SYSTEM])
        _bytes_per_period *= _RECORDSIZE
        return _bytes_per_period

    def get_selective_results(self, columns):
        """
        get results of selective columns in .out-file

        this function is due to its iterative reading slow,
        but has it advantages with out-files with many columns (>1000) and fewer time-steps

        Args:
            columns (list[tuple]): list of column identifier tuple with [(kind, label, variable), ...]

        Returns:
            dict[str, list]: dictionary where keys are the column names ('/' as separator) and values are the list of result values
        """
        n_vars_subcatch = len(self.variables[OBJECTS.SUBCATCHMENT])
        n_vars_node = len(self.variables[OBJECTS.NODE])
        n_vars_link = len(self.variables[OBJECTS.LINK])

        n_subcatch = len(self.labels[OBJECTS.SUBCATCHMENT])
        n_nodes = len(self.labels[OBJECTS.NODE])
        n_links = len(self.labels[OBJECTS.LINK])

        offset_list = list()
        values = dict()

        for kind, label, variable in columns:
            values['/'.join([kind, label, variable])] = list()

            index_kind = OBJECTS.LIST_.index(kind)
            index_variable = self.variables[kind].index(variable)
            item_index = self.labels[kind].index(str(label))
            offset_list.append((2 + index_variable + {
                0: (item_index * n_vars_subcatch),
                1: (n_subcatch * n_vars_subcatch +
                    item_index * n_vars_node),
                2: (n_subcatch * n_vars_subcatch +
                    n_nodes * n_vars_node +
                    item_index * n_vars_link),
                4: (n_subcatch * n_vars_subcatch +
                    n_nodes * n_vars_node +
                    n_links * n_vars_link)
            }[index_kind])*_RECORDSIZE)

        # offset_list = [o*_RECORDSIZE for o in offset_list]
        # cols = list(values.keys())
        # cols_sorted = sorted(cols, key=lambda e: offset_list[cols.index(e)])
        # offset_sorted = sorted(offset_list)
        # iter_label_offset = tuple(zip(cols_sorted, offset_sorted))
        iter_label_offset = tuple(zip(values.keys(), offset_list))

        for period_offset in tqdm(range(self.pos_start_output,  # start
                                        self.pos_start_output + self.n_periods * self._bytes_per_period,  # stop
                                        self._bytes_per_period),
                                  desc=f'{repr(self)}.get_selective_results(n_cols={len(columns)})'):  # step
            # period_offset = self.pos_start_output + period * self.bytes_per_period
            for label, offset in iter_label_offset:
                self._set_position(offset + period_offset)
                values[label].append(self._next_float())

        return values

    def _infer_n_periods(self):
        not_done = True
        period = 0
        while not_done:
            self.fp.seek(self.pos_start_output + period * self._bytes_per_period, SEEK_SET)
            try:
                dt = self._next(dtype='d')
                # print(dt)
                # print(datetime.datetime(1899, 12, 30) + datetime.timedelta(days=dt))
                period += 1
            except:
                not_done = False

        self.n_periods = period - 1



from swmm_api.input_file.macros import calc_slope, get_link_tags, nodes_dict
from swmm_api.input_file.sections import CrossSection, Loss, Vertices
from swmm_api.input_file.sections.link import _Link, Conduit, Orifice, Outlet, Pump, Weir
from swmm_api import SwmmInput
import shape_generator


class SuperConduit(Conduit, CrossSection, Loss, Vertices):
    def __init__(self, inp, conduit, cross_section, loss=None, vertices=None, tag=None):
        self.inp = inp
        Conduit.__init__(self, **conduit.to_dict_())
        CrossSection.__init__(self, **cross_section.to_dict_())
        if loss is not None:
            Loss.__init__(self, **loss.to_dict_())
        if vertices is not None:
            Vertices.__init__(self, **vertices.to_dict_())
        self.tag = tag

        self._nodes = nodes_dict(self.inp)

    @classmethod
    def from_inp(cls, inp, label):
        """

        Args:
            inp (swmm_api.SwmmInput):
            label:

        Returns:

        """
        if label in inp.CONDUITS:
            loss = None
            if label in inp.LOSSES:
                loss = inp.LOSSES[label]

            vertices = None
            if label in inp.VERTICES:
                vertices = inp.VERTICES[label]
            tag = None
            tags = get_link_tags(inp)
            if label in tags:
                tag = tags[label]
            return cls(inp, inp.CONDUITS[label], inp.XSECTIONS[label], loss, vertices, tag)

    @property
    def curve_obj(self):
        if self.Shape == CrossSection.SHAPES.CUSTOM:
            return self.inp.CURVES[self.Curve]

    @property
    def shape_generator(self):
        if self.Shape == CrossSection.SHAPES.CUSTOM:
            return shape_generator.CrossSection.from_curve(self.curve_obj, height=self.Geom1)
        elif self.Shape == CrossSection.SHAPES.IRREGULAR:
            return  # Todo: I don't know how
        elif self.Shape in [CrossSection.SHAPES.RECT_OPEN, CrossSection.SHAPES.RECT_CLOSED]:
            return  # Todo: Rect
        else:
            return shape_generator.swmm_std_cross_sections(self.Shape, height=self.Geom1)

    @property
    def profil_area(self):
        if self.Shape == CrossSection.SHAPES.CUSTOM:
            return self.shape_generator.area_v
        elif self.Shape == CrossSection.SHAPES.IRREGULAR:
            return  # Todo: I don't know how
        elif self.Shape in [CrossSection.SHAPES.RECT_OPEN, CrossSection.SHAPES.RECT_CLOSED]:
            return self.Geom1 * self.Geom2
        else:
            return self.shape_generator.area_v

    @property
    def from_node(self):
        return self._nodes[self.FromNode]

    @property
    def to_node(self):
        return self._nodes[self.ToNode]

    @property
    def slope(self):
        return (self.from_node.Elevation + self.InOffset - (
                self.to_node.Elevation + self.OutOffset)) / self.Length


from swmm_api.input_file.sections import Junction, Storage, Outfall, Coordinate, DryWeatherFlow, Inflow, RainfallDependentInfiltrationInflow, Treatment


class Node(Junction, Storage, Outfall, Coordinate, DryWeatherFlow, Inflow, RainfallDependentInfiltrationInflow, Treatment):
    def __init__(self):
        Junction.__init__(self)
